'''
This module defines some basic constructs used by the GUI, including the
tk_tao_parameter class, which handles displaying and editing a single
parameter in tao
'''
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from tkinter import filedialog
import sys
import os

from pytao.util.parameters import tao_parameter, str_to_tao_param, tao_parameter_dict
from .data_type_list import data_type_list
#from .tao_data_windows import tao_new_data_window, tao_d1_data_window

class tk_tao_parameter():
    '''
    Takes a tao_parameter (defined in parameters.py) and a tk frame,
    and creates an object containing the parameter and appropriate tk widget(s)
    for displaying and modifying the parameter and value
    pipe: the tao_interface object
    data_source: for DAT_TYPE and DAT_TYPE_Z, filters allowed data types
    plot: for DAT_TYPE_Z, the plot where x_axis_type should be checked
    struct_name: for the components of a STRUCT, pass the struct name here
    '''

    def __init__(self, tao_parameter, frame, pipe=0, data_source='', plot='', struct_name=''):
        self.param = tao_parameter
        self.pipe = pipe
        self.sub_wid = None # to be set externally, e.g. by tk_tao_linker
        if struct_name != '':
            struct_name = struct_name + '.'

        if self.param.type == 'DAT_TYPE_Z':
            # Check if operating as ENUM or as DAT_TYPE
            if plot == '': # should never occur
                self.param.type = 'STR' # most fail safe option
            else:
                plot_info = self.pipe.cmd_in('python plot1 ' + plot)
                plot_info = plot_info.splitlines()
                for item in plot_info:
                    if item.find('x_axis_type;') == 0:
                        x_axis_type = item.split(';')[3]
                        break
                    else:
                        x_axis_type = None
                if x_axis_type == 'data':
                    self.param.type = 'DAT_TYPE'
                else:
                    self.param.type = 'ENUM_Z'

        if self.param.is_ignored:
            self.tk_var = tk.StringVar()
            self.tk_var.set(str(self.param.value))
            self.tk_wid = tk.Label(frame, textvariable=self.tk_var)
        elif self.param.type in ['STR', 'INT', 'REAL', 'SPECIES']:
            self.tk_var = tk.StringVar()
            if self.param.value == None:
                self.tk_var.set("")
            else:
                self.tk_var.set(str(self.param.value))
            self.tk_wid = tk.Entry(frame, textvariable=self.tk_var)
        elif self.param.type in ['ENUM', 'ENUM_Z']:
            self.tk_var = tk.StringVar()
            self.tk_var.set(self.param.value)
            if self.param.type == 'ENUM':
                if self.param.prefix == None:
                    options = enum_fetch(self.param.name,pipe)
                else:
                    options = enum_fetch(self.param.prefix + '^' + self.param.name,pipe)
            elif self.param.type == 'ENUM_Z': #added to handle DAT_TYPE_Z
                options = enum_fetch('data_type_z', pipe)
                self.param.type = 'ENUM'
            if options == [""]:
                options = [self.param.value]
            if options[0].find('[FATAL') == 0:
                options = ['[ERROR]']
            if self.param.value == "":
                self.tk_var.set(options[0])
            if self.param.can_vary:
                self.tk_wid = tk.OptionMenu(frame, self.tk_var, *options)
            else:
                self.tk_wid = tk.Label(frame, text=self.param.value)
            # Check for and remove num^ from self.param.name
            self.param.name = self.param.name.split('^')[-1]
        elif self.param.type == 'INUM':
            self.tk_var = tk.StringVar()
            self.tk_var.set(self.param.value)
            if self.param.prefix == None:
                options = inum_fetch(self.param.name,pipe)
            else:
                options = inum_fetch(self.param.prefix + '^' + self.param.name,pipe)
            if options == [""]:
                options = [self.param.value]
            if self.param.value == None:
                self.tk_var.set(options[0])
            self.tk_wid = tk.OptionMenu(frame, self.tk_var, *options)
            # Check for and remove num^ from self.param.name
            self.param.name = self.param.name.split('^')[-1]
        elif self.param.type == 'FILE':
            self.tk_var = tk.StringVar()
            self.tk_var.set(self.param.value)
            if self.tk_var.get() == "":
                self.tk_var.set("Browse...")
            self.tk_wid = tk.Button(
                    frame, textvariable=self.tk_var, command=self.open_file)
        elif self.param.type == 'LOGIC':
            self.tk_var = tk.BooleanVar()
            self.tk_var.set(self.param.value)
            if self.param.can_vary:
                self.tk_wid = tk.Checkbutton(frame, variable=self.tk_var)
            else:
                self.tk_wid = tk.Label(frame, text=str(self.tk_var.get()))
        elif self.param.type == 'REAL_ARR':
            self.tk_var = tk.StringVar()
            val = ""
            for i in range(len(self.param.value)):
                val = val+str(self.param.value[i])
                val = val+';'
            self.tk_var.set(val) #; delimited list of values
            self.tk_wid = tk.Frame(frame) #holds the entries
            self._svar = [] #slave variables
            self._s = [] #slave widgets
            for i in range(len(self.param.value)):
                self._svar.append(tk.StringVar())
                self._svar[i].set(str(self.param.value[i]))
                self._s.append(tk.Entry(self.tk_wid, textvariable=self._svar[i]))
                self._s[i].pack(side="left", expand=1)
                self._svar[i].trace("w", self._update_real_arr)
        elif self.param.type == 'DAT_TYPE':
            self._no_s_refresh = False
            self.tk_var = tk.StringVar() #This is for the entire value
            self.tk_var.set(self.param.value)
            self.tk_wid = tk.Frame(frame) #The "widget" that should be placed by the gui
            self._data_source = data_source
            self._mvar = tk.StringVar() # The master variable
            self._mvar.set((self.tk_var.get()).split('.')[0])
            self._mvar_old = self._mvar.get() # Tracks changes inn self._mvar
            self._m = ttk.Combobox(self.tk_wid, textvariable=self._mvar,
                    values = self._get_dat_types(True), state="readonly")
            self._m.bind("<<ComboboxSelected>>", self._s_refresh)
            self._m.pack(side='left', fill='both', expand=1)
            self._svar = [] #list of slave variables
            self._stype = [] #types of slave widgets (needed for input validation)
            self._s = [] # list of slave widgets
            self._traced = False # tracks whether self._trace_cb exists
            self._s_refresh()
            self._trace_cb = self.tk_var.trace('w', self._fill_widgets)
            self._traced = True
        elif self.param.type == 'DAT_TYPE_E':
            self.tk_var = tk.StringVar()
            self.tk_var.set(self.param.value)
            self.tk_wid = tk.Frame(frame)
            # Sub-widget for the DAT_TYPE portion
            # Parse out the extra info around the DAT_TYPE
            # dat_type_str == part of self.param.value relevant to the dat_type
            dat_type_str = self.param.value
            if self.param.value.find('[') == 0:
                if self.param.value.find('@') != -1:
                    dat_type_str = self.param.value[self.param.value.find('@')+1:]
            if dat_type_str.find('[') != -1:
                dat_type_str = dat_type_str[:dat_type_str.find('[')]
            dat_type_str = self.param.name + ';DAT_TYPE;' + ('T;' if self.param.can_vary else 'F;') + dat_type_str
            self._dat_type = tk_tao_parameter(
                    str_to_tao_param(dat_type_str),
                    self.tk_wid, self.pipe, data_source=data_source)
            # DAT_TYPE_E gets a few extra input fields
            self._uvar = tk.StringVar() # Universe range
            self._u = tk.Entry(self.tk_wid, textvariable=self._uvar)
            self._u.configure(width=10)
            self._evar = tk.StringVar() # Element index
            self._e = tk.Entry(self.tk_wid, textvariable=self._evar)
            self._e.configure(width=10)
            self._cvar = tk.StringVar() # Component
            self._c = tk.Entry(self.tk_wid, textvariable=self._cvar)
            self._c.configure(width=10)
            # Extra labels to put around uni, ele, component
            self._labels = [ tk.Label(self.tk_wid, text='['),
                    tk.Label(self.tk_wid, text=']@'),
                    tk.Label(self.tk_wid, text='['),
                    tk.Label(self.tk_wid, text=']|')]
            # Pack everything
            self._labels[0].pack(side='left')
            self._u.pack(side='left')
            self._labels[1].pack(side='left')
            self._dat_type.tk_wid.pack(side='left', fill='both', expand=1)
            self._labels[2].pack(side='left')
            self._e.pack(side='left')
            self._labels[3].pack(side='left')
            self._c.pack(side='left')
        elif self.param.type == 'STRUCT':
            self.tk_var = tk.StringVar()
            self.tk_var.set('STRUCT')
            self.tk_wid = tk.Frame(frame)
            self.tk_wid.grid_columnconfigure(1, weight=1)
            self._shown = False
            self._m_label = tk.StringVar()
            self._m_label.set("Configure..." if self.param.can_vary else "View...")
            self._m = tk.Button(self.tk_wid, textvariable=self._m_label, command=self._show_hide_struct)
            self._m.grid(row=0, column=0, columnspan=2, sticky='EW')
            self._s = [] # list of tk_tao_parameters
            for component in self.param.value:
                # SPECIAL CASE: x, x2, y, y2 structs
                # do not transfer their name as the prefix
                if self.param.name not in ['x', 'x2', 'y', 'y2']:
                    self._s.append(tk_tao_parameter(component, self.tk_wid, pipe, struct_name=self.param.name))
                else:
                    self._s.append(tk_tao_parameter(component, self.tk_wid, pipe))
        elif self.param.type == 'COMPONENT':
            self.tk_var = tk.StringVar()
            self.tk_var.set(self.param.value)
            if not self.param.can_vary:
                self.tk_wid = tk.Entry(frame, textvariable=self.tk_var)
            else:
                self.tk_wid = tk.Frame(frame)
                self._copts = ['model', 'base', 'design', 'meas', 'ref']
                self._aopts = ['', '+', '-']
                self._handle_block = False
                self.tk_var.trace('w', self._write_comps)
                self._c_refresh()

        if self.param.type not in ['DAT_TYPE', 'DAT_TYPE_E', 'REAL_ARR', 'STRUCT', 'COMPONENT']:
            if (self.param.type != 'FILE') and (not self.param.is_ignored):
                self.tk_wid.config(disabledforeground="black")
        elif self.param.type not in ['DAT_TYPE_E', 'STRUCT']:
            for widget in self._s:
                widget.config(disabledforeground="black")
        elif self.param.type == 'DAT_TYPE_E':
            self._u.config(disabledforeground="black")
            self._e.config(disabledforeground="black")
            self._c.config(disabledforeground="black")
        self.tk_label = tk.Label(frame, text=self.param.name)
        if not self.param.can_vary and not self.param.is_ignored:
            if self.param.type not in ['DAT_TYPE', 'DAT_TYPE_E', 'REAL_ARR', 'STRUCT']:
                self.tk_wid.config(state="disabled")
            elif self.param.type not in ['STRUCT', 'DAT_TYPE_E']:
                for widget in self._s:
                    widget.config(state="disabled")
            elif self.param.type == 'DAT_TYPE_E':
                self._u.config(state='disabled')
                self._e.config(state='disabled')
                self._c.config(state='disabled')
        #Bind info printing
        if pipe != 0:
            if pipe.debug == True:
                self.tk_wid.bind("<Button-3>", self.print_info)

    def print_info(self, *args):
        '''
        Prints diagnostic info about this widget for testing purposes
        '''
        print("Param name: " + self.param.name)
        print("Param type: " + self.param.type)
        print("Param can_vary: " + str(self.param.can_vary))
        print("Param value: " + str(self.param.value))
        print("Widget contents: " + str(self.tk_var.get()))
        print("---------------------")


    def value(self):
        '''
        Returns the value in the input field(s) of self, appropriately typed
        If an invalid value is input, returns None
        '''
        if self.param.type in ['STR', 'ENUM', 'FILE', 'SPECIES']:
            return self.tk_var.get()
        elif self.param.type in ['INT', 'INUM']:
            try:
                return int(self.tk_var.get())
            except:
                return None
        elif self.param.type == 'REAL':
            try:
                return float(self.tk_var.get())
            except:
                return None
        elif self.param.type == 'LOGIC':
            return bool(self.tk_var.get())
        elif self.param.type == 'DAT_TYPE':
            if self._is_valid_dat_type(self.tk_var.get()):
                return self.tk_var.get()
            else:
                return None
        elif self.param.type == 'REAL_ARR':
            output = []
            for i in range(len(self._svar)):
                try:
                    output.append(float(self._svar[i].get()))
                except:
                    output.append(None)
            return output
        elif self.param.type == 'STRUCT':
            d = {}
            for ttp in self._s:
                if ttp.value() != None:
                    d[ttp.param.name] = ttp.value()
            return d

    def copy(self, ttp):
        '''
        Copy the the value(s) of the tk_var(s) of ttp into self
        Does nothing if ttp.param.type != self.param.type
        '''
        if not isinstance(ttp, tk_tao_parameter):
            return
        if self.param.type != ttp.param.type:
            return
        if self.param.type == 'REAL_ARR':
            if len(ttp._svar) != len(self._svar):
                return
            for i in range(len(self._svar)):
                if ttp.value()[i] != None:
                    self._svar[i].set(ttp.value()[i])
        elif self.param.type == 'STRUCT':
            for i in range(len(self._s)):
                if self._s[i].param.name in ttp.value().keys():
                    self._s[i].tk_var.set(ttp.value()[self._s[i].param.name])

    def param_copy(self, tao_param):
        '''
        Copies tao_param.value into self.tk_var (and slaves if appropriate)
        '''
        if not isinstance(tao_param, tao_parameter):
            return
        if self.param.type in ['STR', 'INT', 'REAL', 'SPECIES']:
            if tao_param.value == None:
                self.tk_var.set("")
            else:
                self.tk_var.set(str(tao_param.value))
        elif self.param.type in ['ENUM', 'ENUM_Z']:
            self.tk_var.set(tao_param.value)
        elif self.param.type == 'INUM':
            self.tk_var.set(tao_param.value)
        elif self.param.type == 'FILE':
            self.tk_var.set(tao_param.value)
            if self.tk_var.get() == "":
                self.tk_var.set("Browse...")
        elif self.param.type == 'LOGIC':
            self.tk_var.set(tao_param.value)
            if not self.param.can_vary:
                self.tk_wid.configure(text=str(self.tk_var.get()))
        elif self.param.type == 'REAL_ARR':
            self.tk_var = tk.StringVar()
            val = ""
            for i in range(len(tao_param.value)):
                val = val+str(tao_param.value[i])
                val = val+';'
            self.tk_var.set(val) #; delimited list of values
            # Hack to prevent _update_real_arr from running:
            self.param.type = '~REAL_ARR'
            for i in range(len(tao_param.value)):
                self._svar[i].set(str(tao_param.value[i]))
            # Fix self.param.type
            self.param.type = 'REAL_ARR'
        elif self.param.type == 'DAT_TYPE':
            self.tk_var.set(tao_param.value)
            if self._data_source in ["data", "var"]:
                self._mvar.set((self.tk_var.get()))
                self._mvar_old = self._mvar.get() # Tracks changes in self._mvar
                self._s_refresh()
            else:
                self._mvar.set((self.tk_var.get()).split('.')[0])
                self._mvar_old = self._mvar.get() # Tracks changes in self._mvar
                self._s_refresh()
        elif self.param.type == 'STRUCT':
            for i in range(len(tao_param.value)):
                for j in range(len(self.param.value)):
                    # tao_param.value and self.param.value may not be in the same order
                    if self.param.value[j].name == tao_param.value[i].name:
                        self._s[j].param_copy(tao_param.value[i])
                        break

    def _show_hide_struct(self, event=None, *args):
        '''
        Shows or hides a struct's components as appropriate
        '''
        # Safeguard
        if self.param.type != 'STRUCT':
            return
        if self._shown:
            for ttp in self._s:
                ttp.tk_wid.grid_forget()
                ttp.tk_label.grid_forget()
            self._shown = False
            self._m_label.set("Configure..." if self.param.can_vary else "View...")
        else:
            for i in range(len(self._s)):
                ttp = self._s[i]
                ttp.tk_label.grid(row=i+1, column=0, sticky='W')
                ttp.tk_wid.grid(row=i+1, column=1, sticky='EW')
            self._shown = True
            self._m_label.set("Collapse...")

    def _update_real_arr(self, event=None, *args, **kwargs):
        '''
        Takes a real array's slave variables and puts their data
        into the master tk_var
        '''
        # Safeguard
        if self.param.type != 'REAL_ARR':
            return
        new_val = ""
        for i in range(len(self._svar)):
            # Check that the variable contains a float
            try:
                x = float(self._svar[i].get())
                new_val = new_val + str(float(self._svar[i].get()))
            except:
                # Use zero if not a float
                new_val = new_val + str(0.0)
            new_val = new_val + ';'
        self.tk_var.set(new_val)

    def _get_dat_types(self, filter_data_source=False):
        '''
        Parses the data_type_list (from data_type_list.py) and
        returns a list of allowed master data types
        If filter_data_source is set True, data types whose
        data_source does not match self._data_source are removed
        (no difference if self._data_source is not defined)
        Returns existing d1_data arrays if self._data_source == 'data'
        and existing v1_var arrays if self._data_source == 'var'
        '''
        master_list = []
        # Don't filter if self._data_source = ""
        filter_data_source = False
        if self._data_source not in ['data', 'var']:
            for item in data_type_list:
                # Filter out data_types not allowed for this data source
                if filter_data_source:
                    try:
                        if self._data_source not in item['data_source']:
                            continue
                    except:
                        pass
                item = item['name'].split('<')[0]
                if item != "":
                    if item[-1] == '.':
                        item = item[:-1]
                master_list.append(item)
        elif self._data_source == 'data':
            # Fetch d2 arrays
            d2_arr = self.pipe.cmd_in('python data_d2_array 1').splitlines()
            for d2 in d2_arr:
                d1_list = self.pipe.cmd_in('python data_d1_array ' + d2).splitlines()
                for line in d1_list:
                    master_list.append(d2 + '.' + line.split(';')[3])
        else:
            # Fetch v1 arrays
            var_gen = self.pipe.cmd_in('python var_general').splitlines()
            for line in var_gen:
                master_list.append(line.split(';')[0])
        return master_list

    def _s_refresh(self, event=None, dat_source_swap=False, *args):
        '''
        Clears the existing slave widgets and variables,
        makes new slave widgets and variables, and populates them
        if necessary
        '''
        # Safeguard in case this method gets called for a normal tk_tao_param
        if self.param.type not in ['DAT_TYPE', 'DAT_TYPE_E']:
            return
        if self.param.type == 'DAT_TYPE_E':
            return self._dat_type._s_refresh()
        # Clear the existing slaves
        for item in self._s:
            item.destroy()
        self._svar = []
        self._stype = []
        self._s = []
        self._m.configure(values=self._get_dat_types(True))
        if dat_source_swap:
            if self._data_source in ['data', 'var']:
                self._mvar.set(self.tk_var.get())
            else:
                if self._mvar.get().find('normal.h') == 0:
                    self._mvar.set('normal.h')
                else:
                    self._mvar.set(self._mvar.get().split('.')[0])
            self._mvar_old = self._mvar.get()

        try:
            if self._data_source in ['data' , 'var']:
                m_ix = 0
            else:
                m_ix = (self._get_dat_types()).index(self._mvar.get())
        except ValueError:
            m_ix = 0

        # Make new slave widgets
        dat_dict = data_type_list[m_ix]
        p_start = dat_dict['name'].find('<') - 1
        if p_start != -2:
            slave_params = (dat_dict['name'][p_start:]).split('.')[1:]
        else:
            slave_params = []
        k = 0 #loop counter
        for p in slave_params:
            self._svar.append(tk.StringVar())
            self._stype.append(p)
            if p.find('<enum') != -1: #Enums
                self._s.append(tk.OptionMenu(self.tk_wid, self._svar[k],
                    *dat_dict[p]))
                self._svar[k].set(dat_dict[p][0]) # Default to first list item
            elif p.find('<digit:') != -1: #Digit dropdown box
                p = p.split(':')[1]
                p = p.split('>')[0] #p is now "low-high"
                low, high = p.split('-')
                low = int(low)
                high = int(high)
                digits = list(range(low, high+1))
                self._s.append(tk.OptionMenu(self.tk_wid, self._svar[k],
                    *digits))
                self._svar[k].set(digits[0])
            elif p.find('<str>') != -1: #Strings
                self._s.append(tk.Entry(self.tk_wid, textvariable=self._svar[k]))
            elif p.find('<digits') != -1: #Fixed length int
                p = p.split('s')[1]
                p = p.split('>')[0]
                length = int(p)
                self._s.append(tk.Entry(self.tk_wid, textvariable=self._svar[k]))
            elif p.find('<int>') != -1: #Integer
                self._s.append(tk.Entry(self.tk_wid, textvariable=self._svar[k]))
            elif p.find('<real>') != -1: #Float
                self._s.append(tk.Entry(self.tk_wid, textvariable=self._svar[k]))
            self._svar[k].trace("w", self._update_tk_var)
            k = k+1

        # Set the slave variables appropriately
        #current_mvar = (self.tk_var.get()).split('<')[0]
        #if current_mvar != "":
        #  if current_mvar[-1] == '.':
        #        current_mvar = current_mvar[:-1]
        #current_mvar = current_mvar.split('.')[0]
        if self._mvar.get() == self._mvar_old:
            for k in range(len(self._svar)):
                # Special case: velocity -> velocity.
                if self.tk_var.get() == 'velocity':
                    self._svar[k].set("velocity.".split('.')[k+1])
                else:
                    self._svar[k].set((self.tk_var.get()).split('.')[k+1])
        else: # Update self.tk_var if self._mvar has changed
            self._update_tk_var()
            self._mvar_old = self._mvar.get()

        # Pack the new slaves
        for k in range(len(self._s)):
            self._s[k].pack(side="left", fill="both")

    def _update_tk_var(self, event=None, *args, **kwargs):
        '''
        Updates self.tk_var with the current contents of
        self._mvar and self._svar
        '''
        #print("_update_tk_var called for " + self.param.name)
        #print("param.type = " + self.param.type)
        #print("self._data_soure = " + self._data_source)
        if self.param.type == 'DAT_TYPE_E':
            self._dat_type._update_tk_var()
            self.tk_var.set('['+self._uvar.get()+']@'
                    +self._dat_type.tk_var.get()
                    +'['+self._evar.get()+']|'
                    +self._cvar.get())
        if self.param.type != 'DAT_TYPE':
            return
        #if self._data_source not in ['lat', 'beam']:
        #    return
        new_tk_var = self._mvar.get()
        #print("new_tk_var = " + new_tk_var)
        for k in range(len(self._svar)):
            # Input validation (TODO)
            p = self._stype[k]
            if p.find('<digits') != -1:
                p = p.split('s')[1]
                p = p.split('>')[0]
                length = int(p)
                #Check if the last character typed was a digit
                if len(self._svar[k].get()) != 0:
                    try:
                        x = int(self._svar[k].get()[-1])
                    except:
                        self._svar[k].set(self._svar[k].get()[:-1])
                #Prevent length from being greater than 6 digits
                self._svar[k].set((self._svar[k].get())[-1*length:])
                #Write the corrent number of digits (pad with zeros)
                try: #see if self._svar[k] is an int
                    x = int(self._svar[k].get())
                    x = len(self._svar[k].get())
                    if x < 6:
                        new_tk_var = new_tk_var + '.' + (6-x) * '0' + self._svar[k].get()
                    else:
                        new_tk_var = new_tk_var + '.' + self._svar[k].get()
                except ValueError:
                    new_tk_var = new_tk_var + '.' + length * '0'
            elif p == '<int>':
                try: #see if self._svar[k] is an int
                    x = int(self._svar[k].get())
                    new_tk_var = new_tk_var + '.' + self._svar[k].get()
                except ValueError:
                    new_tk_var = new_tk_var + '.' + '0'
            elif p == '<real>':
                try: #see if self._svar[k] is an float
                    x = float(self._svar[k].get())
                    new_tk_var = new_tk_var + '.' + self._svar[k].get()
                except ValueError:
                    new_tk_var = new_tk_var + '.' + '0'
            else:
                new_tk_var = new_tk_var + '.' + self._svar[k].get()
            #print("new_tk_var = " + new_tk_var)
        # Special case: velocity. -> velocity
        if new_tk_var == "velocity.":
            new_tk_var = "velocity"
        # Un-trace tk_var to prevent repeatedly running this method
        self._no_s_refresh = True
        #print("new_tk_var = " + new_tk_var)
        self.tk_var.set(new_tk_var)
        # Re-trace tk_var
        self._no_s_refresh = False
        #self._trace_cb = self.tk_var.trace('w', self._fill_widgets)
        #self._traced = True

    def _is_valid_dat_type(self, x):
        '''
        Returns True if the string x is a valid data_type and False otherwise
        '''
        if not isinstance(x, str):
            return False
        # Special case: normal.h
        if x.find('normal.h') == 0:
            x = ['normal.h'] + x[8:].split('.')
        else:
            x = x.split('.')
        try:
            m_ix = (self._get_dat_types()).index(x[0])
        except ValueError:
            return False

        # Check what subvalues are allowed
        dat_dict = data_type_list[m_ix]
        p_start = dat_dict['name'].find('<') - 1
        if p_start != -2:
            slave_params = (dat_dict['name'][p_start:]).split('.')[1:]
        else:
            slave_params = []
        # Reject if length of slave_params and x don't match up
        if len(slave_params) != len(x)-1:
            return False

        # Check each subvalue
        k = 0 #loop counter
        for p in slave_params:
            if p.find('<enum') != -1: #Enums
                if x[k+1] not in dat_dict[p]:
                    return False
            elif p.find('<digit:') != -1: #Digit dropdown box
                try:
                    x[k+1] = int(x[k+1])
                except ValueError:
                    return False
                p = p.split(':')[1]
                p = p.split('>')[0] #p is now "low-high"
                low, high = p.split('-')
                low = int(low)
                high = int(high)
                if (x[k+1] < low) | (x[k+1] > high):
                    return False
            elif p.find('<str>') != -1: #Strings
                if len(x[k+1]) == 0:
                    return False
            elif p.find('<digits') != -1: #Fixed length int
                try:
                    junk_var = int(x[k+1]) #don't need the value
                except ValueError:
                    return False
                p = p.split('s')[1]
                p = p.split('>')[0]
                length = int(p)
                if len(x[k+1]) != length:
                    return False
            elif p.find('<int>') != -1: #Integer
                try:
                    junk_var = int(x[k+1]) #don't need the value
                except ValueError:
                    return False
            elif p.find('<real>') != -1: #Float
                try:
                    junk_var = float(x[k+1]) #don't need the value
                except ValueError:
                    return False
            k = k+1
        return True # All tests passed

    def _fill_widgets(self, event=None, *args):
        '''
        Runs self._s_refresh() and then fills the slave widgets
        with data from self.tk_var, but only if self.is_valid_dat_type(self.tk_var.get()) is True
        '''
        if self.param.type != 'DAT_TYPE':
            return
        if not self._is_valid_dat_type(self.tk_var.get()):
            return

        # Refresh slave widgets
        self._mvar.set((self.tk_var.get()).split('.')[0])
        self._mvar_old = self._mvar.get() # gets the slave variables to fill
        if self._no_s_refresh:
            return
        self._s_refresh()

    def _has_ele(self, *args):
        '''
        Returns True if self.tk_var.get() can take associated ele names
        If pipe == 0, returns True by default
        If self.param.type != DAT_TYPE, returns false
        '''
        if self.param.type == 'DAT_TYPE_E':
            return self._dat_type._has_ele()
        if self.param.type != 'DAT_TYPE':
            return False
        if self.pipe == 0:
            return True

        cmd_string = 'python datum_has_ele '
        cmd_string += self.tk_var.get()
        if self.pipe.cmd_in(cmd_string) == 'no':
            return False
        else:
            return True

    def _has_s_offset(self, *args):
        '''
        Returns True if self.tk_var.get() can take an s_offset
        Returns False if self.param.type != DAT_TYPE
        '''
        if self.param.type == 'DAT_TYPE_E':
            return self._dat_type._has_s_offset()
        if self.param.type != 'DAT_TYPE':
            return False

        # Locate the correct dictionary in data_type_list
        for data_dict in data_type_list:
            if data_dict['name'].find(self._mvar.get()) != -1:
                found=True
                break
            else:
                found=False

        if not found:
            return True #default

        return data_dict['s_offset']

    def _c_refresh(self):
        '''
        Reads the contents of self.tk_var and creates the proper number of
        OptionMenu widgets
        '''
        if self.param.type != 'COMPONENT':
            return
        for child in self.tk_wid.winfo_children():
            child.destroy()
        # parse self.tk_var
        val = self.tk_var.get()
        def parse_comps(x):
            '''
            Recursive function to read off one term in the component string.
            Returns a list in the form [comp1, +/-, comp2, +/-, ..., compN]
            '''
            x = x.strip()
            pix = x.find('+')
            if pix == -1:
                pix = len(x)
            mix = x.find('-')
            if mix == -1:
                mix = len(x)
            if pix < mix:
                a = '+'
            elif mix < pix:
                a = '-'
            else:
                a = None
            if a == None:
                return [x]
            else:
                return [x[:x.find(a)]] + [a] + parse_comps(x[x.find(a)+1:])
        val = parse_comps(val) #len(val) is odd
        # create widgets
        self._s = [] # widgets
        self._svar = [] # tk variables
        for i in range(len(val)+1):
            if i%2 == 0: #component
                opts = self._copts
            else: #+/-
                opts = self._aopts
            self._svar.append(tk.StringVar())
            self._svar[-1].set(val[i] if i<len(val) else "")
            self._s.append(tk.OptionMenu(self.tk_wid, self._svar[-1], *opts))
            self._s[-1].pack(side='left')
            self._svar[-1].trace('w', self._write_comps)

    def _write_comps(self, *args):
        '''
        Writes the contents of self._svar into self.tk_var (for COMPONENT
        parameters only), then calls self._c_refresh()
        '''
        if self._handle_block == True: # prevent this function from triggering itself
            return
        if self.param.type != 'COMPONENT':
            return
        val = ""
        for v in self._svar:
            if v.get() != "":
                val += v.get()
            else:
                break
        self._handle_block = True
        self.tk_var.set(val)
        #print(self.param.name + 'set to ' + val)
        self._handle_block = False
        self._c_refresh()

    def open_file(self):
        '''
        Provides a dialog window for selecting a file
        '''
        filename = filedialog.askopenfilename(title = "Select " + self.param.name)
        #Only set if the user has selected an actual file
        if isinstance(filename, str) & (filename != ""):
            self.tk_var.set(filename)
        else:
            self.tk_var.set("Browse...")

    # This makes it easier to search lists of tk_tao_parameters
    def __eq__(self, name):
        return self.param.name == name

#-----------------------------------------------------------------
class d1_data_list_entry():
    '''Creates various tk widgets to display attributes of a single datum.
    Takes one line of output from python data_d_array as input.
    '''
    def __init__(self, master, d_list):
        d_list = d_list.split(';')
        self.tk_wids = [] #Holds all the widgets
        self.tk_tao_params = {} #Holds the tk_tao_parameters for non-label items
        self.index = d_list[0]
        for i in range(6):
            self.tk_wids.append(tk.Label(master, text=d_list[i]))

        #Meas value
        meas_value = tao_parameter("meas_value", "REAL", "T", d_list[6])
        self.tk_tao_params["meas_value"] = tk_tao_parameter(meas_value, master)
        self.tk_wids.append(self.tk_tao_params["meas_value"].tk_wid)

        #Model value
        model_value = tao_parameter("model_value", "REAL", "F", d_list[7])
        self.tk_tao_params["model_value"] = tk_tao_parameter(model_value, master)
        self.tk_wids.append(self.tk_tao_params["model_value"].tk_wid)

        #Design value
        design_value = tao_parameter("design_value", "REAL", "F", d_list[8])
        self.tk_tao_params["design_value"] = tk_tao_parameter(design_value, master)
        self.tk_wids.append(self.tk_tao_params["design_value"].tk_wid)

        #Useit_opt and Useit_plot
        useit_opt = tao_parameter("useit_opt", "LOGIC", "F", d_list[9])
        self.tk_tao_params["useit_opt"] = tk_tao_parameter(useit_opt, master)
        self.tk_wids.append(self.tk_tao_params["useit_opt"].tk_wid)

        useit_plot = tao_parameter("useit_plot", "LOGIC", "F", d_list[10])
        self.tk_tao_params["useit_plot"] = tk_tao_parameter(useit_plot, master)
        self.tk_wids.append(self.tk_tao_params["useit_plot"].tk_wid)

        #Good User
        good_user = tao_parameter("good_user", "LOGIC", "T", d_list[11])
        self.tk_tao_params["good_user"] = tk_tao_parameter(good_user, master)
        self.tk_wids.append(self.tk_tao_params["good_user"].tk_wid)

        #Weight
        weight = tao_parameter("weight", "REAL", "T", d_list[12])
        self.tk_tao_params["weight"] = tk_tao_parameter(weight, master)
        self.tk_wids.append(self.tk_tao_params["weight"].tk_wid)

class v1_var_list_entry():
    '''
    Creates the widgets needed to display a single variable held in a v1_var array
    v_list should be one row of output from
    python var_v_array [variable name]
    '''
    def __init__(self, master, v_list):
        v_list = v_list.split(';')
        self.tk_wids = [] #Holds all the widgets
        self.tk_tao_params = {} #Holds the tk_tao_parameters for non-label items

        self.index = v_list[0]
        for i in range(2):
            self.tk_wids.append(tk.Label(master, text=v_list[i]))

        # Meas value
        meas_value = tao_parameter("meas_value", "REAL", "T", v_list[2])
        self.tk_tao_params["meas_value"] = tk_tao_parameter(meas_value, master)
        self.tk_wids.append(self.tk_tao_params["meas_value"].tk_wid)

        # Model value
        model_value = tao_parameter("model_value", "REAL", "F", v_list[3])
        self.tk_tao_params["model_value"] = tk_tao_parameter(model_value, master)
        self.tk_wids.append(self.tk_tao_params["model_value"].tk_wid)

        # Design value
        design_value = tao_parameter("design_value", "REAL", "F", v_list[4])
        self.tk_tao_params["design_value"] = tk_tao_parameter(design_value, master)
        self.tk_wids.append(self.tk_tao_params["design_value"].tk_wid)

        # Useit_opt
        useit_opt = tao_parameter("useit_opt", "LOGIC", "F", v_list[5])
        self.tk_tao_params["useit_opt"] = tk_tao_parameter(useit_opt, master)
        self.tk_wids.append(self.tk_tao_params["useit_opt"].tk_wid)

        # good_user
        good_user = tao_parameter("good_user", "LOGIC", "T", v_list[6])
        self.tk_tao_params["good_user"] = tk_tao_parameter(good_user, master)
        self.tk_wids.append(self.tk_tao_params["good_user"].tk_wid)

        # Weight
        weight = tao_parameter("weight", "REAL", "T", v_list[7])
        self.tk_tao_params["weight"] = tk_tao_parameter(weight, master)
        self.tk_wids.append(self.tk_tao_params["weight"].tk_wid)

#-----------------------------------------------------------------

def enum_fetch(enum,pipe):
    '''
    Takes the name of an enum variable and returns a list of its allowed values
    using the given pipe
    '''
    if pipe != 0:
        list_string = pipe.cmd_in("python enum " + enum)
        option_list = list_string.splitlines()
        for i in range(len(option_list)):
            sc = option_list[i].find(';')
            option_list[i] = option_list[i][sc+1:]
        # Special case: ele shapes
        # Have to add in custom shapes
        if enum == "shape^shape":
            custom_shapes = pipe.cmd_in("python shape_pattern_list")
            custom_shapes = custom_shapes.splitlines()
            for i in range(len(custom_shapes)):
                custom_shapes[i] = custom_shapes[i].split(';')[0].strip()
            # Insert in custom pattern slot if possible
            if "Pattern:<pattern-name>" in option_list:
                ix = option_list.index("Pattern:<pattern-name>")
                option_list = option_list[:ix] + custom_shapes + option_list[ix+1:]
            else:
                option_list = option_list + custom_shapes
    else:
        option_list = ["TAO NOT STARTED"]
    if option_list == []:
        option_list = [""]
    return option_list


def inum_fetch(inum,pipe):
    '''
    Takes the name of an inum variable and returns a list of its allowed values
    using the given pipe
    '''
    if pipe != 0:
        list_string = pipe.cmd_in("python inum " + inum)
        option_list = list_string.splitlines()
    else:
        option_list = ["TAO NOT STARTED"]
    if option_list == []:
        option_list = [""]
    return option_list

def tk_tao_linker(tk_tao_list):
    '''
    Takes a list of tk_tao_parameters as input and returns a list of tk_tao_parameters
    The output list will have parameters marked with I either removed or
    added to their appropriate parent parameter's tk_wid
    '''
    # determined from ttp.is_ignored:
    sub_widget_names = [] # the names
    sub_widgets = [] # the widgets
    sub_widget_indices = [] # the indices in tk_tao_list
    # First pass: collect sub_widgets
    for i in range(len(tk_tao_list)):
        ttp = tk_tao_list[i]
        if ttp.param.is_ignored:
            # add parentheses around value
            ttp.tk_var.set('(' + ttp.tk_var.get() + ')')
            sub_widget_names.append(ttp.param.name)
            sub_widgets.append(ttp.tk_wid)
            sub_widget_indices.append(i)
    # Second pass: link sub_widgets to their masters
    for ttp in tk_tao_list:
        sub = ttp.param.sub_param
        if sub in sub_widget_names:
            ttp.sub_wid = sub_widgets[sub_widget_names.index(sub)]
    # Third pass: remove ignored widgets
    output = []
    sub_widget_indices.append(len(tk_tao_list)) #fixes the loop below
    for i in range(len(sub_widget_indices)):
        ix = sub_widget_indices[i]
        if i==0:
            output += tk_tao_list[:ix]
        else:
            last_ix = sub_widget_indices[i-1]
            output += tk_tao_list[last_ix+1:ix]
    return output



import tkinter as tk
import os

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.backend_tools import *
from matplotlib import rcParams
from matplotlib.widgets import Slider
from matplotlib.backend_bases import ToolContainerBase
from matplotlib.backends._backend_tk import FigureManagerTk, ToolbarTk
from matplotlib.backend_managers import ToolManager

from pytao.util.parameters import *

#---------------------------------------------
# Note: Help and recalculate buttons are commented out since later version of matplotlib/tkinter will not
# work with absolute path names. It would be good to fix this at some point...

class taotoolbar(NavigationToolbar2Tk):
    def __init__(self,canvas_,parent_,width_,GUI_DIR_):
        self.basedir = os.path.join(rcParams['datapath'], 'images')
        self.GUI_DIR = GUI_DIR_
        self.toolitems = (
            ('Home', 'Reset original view', 'home', 'home'),
            ('Back', 'Back to previous view', 'back', 'back'),
            ('Forward', 'Forward to next view', 'forward', 'forward'),
            ('Pan', 'Pan axes with left mouse, zoom with right mouse or scroll wheel', 'move', 'pan'),
            ('Zoom', 'Zoom to rectangle', 'zoom_to_rect', 'zoom'),
##            ('Redraw','Recalculate points',self.GUI_DIR+'/assets/recalculate','redraw'),
            ('Slider','Width slider','subplots','slider'),
            (None,None,None,None),
            ('Save', 'Save image of figure', 'filesave', 'save_figure'),
            (None,None,None,None),
##            ('Help','Graph help',self.GUI_DIR+'/assets/help','help'),
            )
        self.parent = parent_
        self.canvas = canvas_
        self.width = width_ #width used with slider tool
        self.axes_list = self.canvas.figure.get_axes()
        self.templateDict = self.parent.root.placed
        self.graph_list = self.parent.fig_info[18]
        self.template = self.templateDict[self.parent.region] #corresponding template for a region in tao
        self.templateGraphList = []
        for i in self.graph_list:
            self.templateGraphList.append(self.template + '.' + i.split('.')[1])
        self.gRangeList = [] #list of original x min, x max, y min, and y max for each graph
        try:
            for i in self.templateGraphList:
                gInfo=self.parent.pipe.cmd_in('python plot_graph '+i,no_warn = True).splitlines()
                gRange = []
                gInfoDict = {}
                for i in range(len(gInfo)):
                    gInfoDict[gInfo[i].split(';')[0]]=str_to_tao_param(gInfo[i])
                gRange.append(gInfoDict['x'].get_component('min'))
                gRange.append(gInfoDict['x'].get_component('max'))
                gRange.append(gInfoDict['y'].get_component('min'))
                gRange.append(gInfoDict['y'].get_component('max'))
                self.gRangeList.append(gRange)
        except IndexError: #incase number of graphs changes, like in wave analysis
            pass
        NavigationToolbar2Tk.__init__(self,canvas_,parent_)

    cid = 'none' #connection id for scroll wheel
    cidKeyP = 'none' #connection id for key press
    cidKeyR = 'none' #connection id for key release

    cur_ax = 'none' #axes instance that the mouse is currently over or 'none'
    xzoom = True #allow for x axis scaling
    yzoom = True #allow for y axis scaling



    def enter_axes(self,event):
        '''handles the mouse entering axes'''
        if event.inaxes is not None:
            self.cur_ax = event.inaxes


    def leave_axes(self,event):
        '''handles the mouse leaving axes'''
        if event.inaxes is None:
            self.cur_ax = 'none'

    def onKeyPress(self,event):
        '''handles key presses'''
        if event.key == 'x':
            self.xzoom = True
            self.yzoom = False
        if event.key == 'y':
            self.xzoom = False
            self.yzoom = True

    def onKeyRelease(self,event):
        '''handles releasing key presses'''
        self.xzoom = True
        self.yzoom = True



    def zoom_factory(self,axes,canv,event,on = False,cid='none',base_scale = 1.5):
        '''controls connections for scroll wheel zooming'''

        fig = canv[0].get_figure() # get the figure of interest

        enter = fig.canvas.mpl_connect('axes_enter_event', self.enter_axes)
        leave = fig.canvas.mpl_connect('axes_leave_event', self.leave_axes)

        self.cidKeyP = fig.canvas.mpl_connect('key_press_event',self.onKeyPress)
        self.cidKeyR = fig.canvas.mpl_connect('key_release_event',self.onKeyRelease)

        def zoom_fun(event):
            '''changes graph axes if scroll wheel is used'''
            if self.cur_ax != 'none':
                try:
                    ax = self.cur_ax

                    cur_xlim = ax.get_xlim()
                    cur_ylim = ax.get_ylim()
                    #get the current x and y limits

                    xdata = event.xdata #get event x location
                    ydata = event.ydata #get event y location
                    x_left = xdata - cur_xlim[0]
                    x_right = cur_xlim[1] - xdata
                    y_top = ydata - cur_ylim[0]
                    y_bottom = cur_ylim[1] - ydata
                    if event.button == 'up': # dealwith zoom in
                        scale_factor = 1/base_scale
                    elif event.button == 'down': #deal with zoom out
                        scale_factor = base_scale
                    else:
                        scale_factor = 1
                    if self.xzoom == True: #new x limits
                        ax.set_xlim([xdata - x_left*scale_factor,xdata + x_right*scale_factor])
                    if self.yzoom == True: #new y limits
                        ax.set_ylim([ydata - y_top*scale_factor,ydata + y_bottom*scale_factor])
                    self.canvas.draw_idle()
                except TypeError:
                    pass #scrolling outside of graph
        # attach the call backs
        if on == True and cid == 'none':
            self.cid=fig.canvas.mpl_connect('scroll_event',zoom_fun)

        if on != True:
            fig.canvas.mpl_disconnect(cid)
            self.cid='none'

        if self._nav_stack() is None: #make home button work with scroll wheel zoom
            self.push_current()

        #return the function
        return zoom_fun



    def home(self, *args):
        """Restore the original view."""
        try:
            for i in range(len(self.graph_list)):
                self.parent.pipe.cmd_in('set graph '+self.graph_list[i]+' x.min = '+str(self.gRangeList[i][0]),no_warn = True)
                self.parent.pipe.cmd_in('set graph '+self.graph_list[i]+' x.max = '+str(self.gRangeList[i][1]),no_warn = True)
                self.parent.pipe.cmd_in('set graph '+self.graph_list[i]+' y.min = '+str(self.gRangeList[i][2]),no_warn = True)
                self.parent.pipe.cmd_in('set graph '+self.graph_list[i]+' y.max = '+str(self.gRangeList[i][3]),no_warn = True)
        except IndexError:
            pass
        self.parent.refresh()
        self._nav_stack.home()
        self.set_history_buttons()
        self._update_view()



    def pan(self, *args):
        """Activate the pan/zoom tool. pan with left button, zoom with right or with with scroll wheel"""
        #set the pointer icon and button press funcs to the appropriate callbacks
        #default matplotlib pan tool, but with added scroll wheel zooming in pan/zoom mode
        if self._active == 'PAN':
            self._active = None
        else:
            self._active = 'PAN'
        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)
            self.mode = ''

        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
            self.mode = ''

        if self._active:

            self._idPress = self.canvas.mpl_connect(
                'button_press_event', self.press_pan)
            self._idRelease = self.canvas.mpl_connect(
                'button_release_event', self.release_pan)
            self.mode = 'pan/zoom'
            scale_canv = self.canvas
            scale = self.zoom_factory(self,scale_canv.figure.get_axes(),scale_canv,True,self.cid)
            self.canvas.widgetlock(self)
        else:
            self.canvas.widgetlock.release(self)
            scale_canv = self.canvas
            scale = self.zoom_factory(self,scale_canv.figure.get_axes(),scale_canv,False,self.cid)
        for a in self.canvas.figure.get_axes():
            a.set_navigate_mode(self._active)

        self.set_message(self.mode)



    def help(self):
        '''help tool which opens a tk window with instructions on graph interaction'''
        image = None

        win = tk.Toplevel(self.parent)

        # Title string
        title = 'Graph Help'

        #Help items and descriptions
        help_items = ['Home',  'Back', 'Forward', 'Pan/Zoom', 'Zoom Rectangle', 'Recalculate Points', 'Width Slider', 'Save', '<Double Click>']
        help_descrips = [
            "Returns to original view with original points, shortcuts are 'h' or 'r'.",
            "Returns to previous view, shortcuts are 'c' or 'left arrow'.",
            "Undoes the last back command, shortcuts are 'v' or 'right arrow'.",
            """Toggles panning by left clicking and dragging, and zooming by using the scroll wheel or by right clicking and dragging. 
Holding 'x' restricts panning and zooming to the x axis, holding 'y' restricts panning and zooming to the y axis, and holding control maintains the aspect ratio.""",
            """Toggles zooming by left clicking and dragging to select the new window. 
Holding 'x' restricts zooming to the x axis, holding 'y' restricts zooming to the y axis, and holding control maintains the aspect ratio.""",
            "Recalculates points to better fit the current window size",
            "Makes a slider that changes the size of a lat layout if one is drawn below a graph, or changes width of elements in a floor plan.",
            "Saves the current figure as an image file, shortcut is 'ctrl + s'.",
            "With Lat_Layout or Floor_Plan: Double click on Shape to open element parameter window."]

        wl = 600
        tk.Label(win, text=title).grid(row=0, column=0, columnspan=2, sticky='EW')
        for i in range(len(help_items)):
            tk.Label(win, text=help_items[i]).grid(row=2*i+1, column=0, sticky='NW', padx = 10)
            tk.Label(win, text='').grid(row=2*i+2, column=0, sticky='W')
            tk.Label(win, text=help_descrips[i], wraplength=wl, justify='left').grid(row=2*i+1, column=1, sticky='W')
            tk.Label(win, text='', wraplength=wl, justify='left').grid(row=2*i+2, column=1, sticky='W')
        win.grid_columnconfigure(1, weight=1)



    def redraw(self):
        '''forces recalculation of points in window'''

        for i in range(len(self.graph_list)):
            xRange = self.axes_list[i].get_xlim()
            yRange = self.axes_list[i].get_ylim()
            self.parent.pipe.cmd_in('set graph '+self.graph_list[i]+' x.min = '+str(xRange[0]),no_warn = True)
            self.parent.pipe.cmd_in('set graph '+self.graph_list[i]+' x.max = '+str(xRange[1]),no_warn = True)
            self.parent.pipe.cmd_in('set graph '+self.graph_list[i]+' y.min = '+str(yRange[0]),no_warn = True)
            self.parent.pipe.cmd_in('set graph '+self.graph_list[i]+' y.max = '+str(yRange[1]),no_warn = True)

        self.parent.refresh()



    def slider(self):
        '''opens a new matplotlib figure with a slider that adjusts the size of a lat layout if one is present or adjusts the width of floor plan elements if a floor plan is present'''
        rcParams['toolbar'] = 'None' #hides toolbar on figure with slider
        slidefig = plt.figure(figsize=[4,.7])
        slideax = slidefig.add_subplot(1,1,1)
        plt.title('Element Width Slider')
        width_slider = Slider(slideax, 'width', 0, 2, self.width) #element width slider
        slidefig.tight_layout(pad=.5)
        slidefig.show()
        def update_slider(width):
            for i in range(len(self.graph_list)):
                xRange = self.axes_list[i].get_xlim()
                yRange = self.axes_list[i].get_ylim()
            self.parent.refresh(width=width_slider.val)
            for i in range(len(self.graph_list)):
                self.axes_list[i].set_xlim(xRange)
                self.axes_list[i].set_ylim(yRange)
        width_slider.on_changed(update_slider) #call update when slider moves






import subprocess
from hashlib import blake2b
import numpy as np
import json
import os

"""
General tools, not specific to Tao


"""


def full_path(path):
    """
    Helper function to expand enviromental variables and return the absolute path
    """
    return os.path.abspath(os.path.expandvars(path))


class NpEncoder(json.JSONEncoder):
    """
    See: https://stackoverflow.com/questions/50916422/python-typeerror-object-of-type-int64-is-not-json-serializable/50916741
    """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)

def fingerprint(keyed_data, digest_size=16):
    """
    Creates a cryptographic fingerprint from keyed data. 
    Used JSON dumps to form strings, and the blake2b algorithm to hash.
    
    """
    h = blake2b(digest_size=16)
    for key in sorted(keyed_data.keys()):
        val = keyed_data[key]
        s = json.dumps(val, sort_keys=True, cls=NpEncoder).encode()
        h.update(s)
    return h.hexdigest()  


def execute(cmd, cwd=None):
    """
    
    Constantly print Subprocess output while process is running
    from: https://stackoverflow.com/questions/4417546/constantly-print-subprocess-output-while-process-is-running
    
    # Example usage:
        for path in execute(["locate", "a"]):
        print(path, end="")
        
    Useful in Jupyter notebook
    
    """
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True, cwd=cwd)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line 
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)
        
# Alternative execute
def execute2(cmd, timeout=None, cwd=None):
    """
    Execute with time limit (timeout) in seconds, catching run errors. 
    """
    
    output = {'error':True, 'log':''}
    try:
        p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, timeout = timeout, cwd=cwd)
      #  p = subprocess.run(' '.join(cmd), shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, timeout = timeout)
        output['log'] = p.stdout
        output['error'] = False
        output['why_error'] =''
    except subprocess.TimeoutExpired as ex:
        output['log'] = ex.stdout+'\n'+str(ex)
        output['why_error'] = 'timeout'
    except:
        output['log'] = 'unknown run error'
        output['why_error'] = 'unknown'
    return output
        



from setuptools import setup, find_packages
from os import path, environ
from codecs import open

import versioneer

cur_dir = path.abspath(path.dirname(__file__))

with open(path.join(cur_dir, 'requirements.txt'), 'r') as f:
    requirements = f.read().split()

with open(path.join(cur_dir, 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pytao',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(),  
    package_dir={'pytao':'pytao'},
    url='https://www.classe.cornell.edu/bmad/tao.html',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    include_package_data=True,
    python_requires='>=3.6'
)



"""
Class for piping in commands to Tao from python and for grabbing Tao output.

This module needs the pexpect module which can be downloaded from:
  <http://sourceforge.net/projects/pexpect/>

Example:
  import tao_pipe
  pipe = tao_pipe.tao_io("-lat my_lat.bmad")   # Init
  pipe.cmd("show uni")                         # Issue a command & print output at terminal.
  tao_output = pipe.cmd_in("show uni")         # Get the output of a command. No terminal output.
"""

import pexpect
import os
import string
import sys

class tao_io:

  #-----------------------------------------------------------------
  # init_args  = Startup command line args.
  # tao_exe    = Tao executable file name including path.

  def __init__(self, init_args = '', tao_exe = '', timeout = 600, expect_str = 'Tao>'):
    if tao_exe == '': tao_exe = '$ACC_EXE/tao'
    init_string = tao_exe + ' ' + init_args
    init_string = string.Template(init_string).substitute(os.environ) # Expand environmental variables.

    if sys.version[0] == '2':                        # Python 2?
      if expect_str == 'Tao>': expect_str = u'Tao>'  # Python 2 unicode compatability.
      self.pipe = pexpect.spawn (init_string, timeout = timeout)
    else:
      self.pipe = pexpect.spawn (init_string, timeout = timeout, encoding = 'utf-8', codec_errors='replace')

    self.expect_str = expect_str
    self.is_open = True

    try:
      self.pipe.expect (expect_str)
    except pexpect.EOF:
      self.pipe.close
      self.is_open = False

    print (self.pipe.before)

  #-----------------------------------------------------------------
  # tao_io.cmd_in method sends a command to Tao and returns the output string

  def cmd_in (self, cmd_str):

    self.cmd_str = cmd_str

    if not self.is_open:
      print ('Not connected to Tao...')
      return ''

    self.pipe.sendline (cmd_str)
    try:
      self.pipe.expect (self.expect_str)
    except pexpect.EOF:
      self.pipe.close
      self.is_open = False

    # pipe.before is a multiline string. The first line is the input command string.
    # This first line is striped off before it is returned

    return self.pipe.before.partition('\n')[2].strip()

  #-----------------------------------------------------------------
  # tao_io.cmd method calls tao_io.cmd and prints the output, including
  # the command and Tao prompt, to the terminal.
  # Note: self.pipe.after will always be the expect string which is generally 'Tao>'

  def cmd (self, cmd_str):
    self.cmd_in(cmd_str)
    if self.is_open:
      print (self.pipe.after + self.pipe.before)
    else:
      print ('Not connected to Tao...')



'''
This module defines the tao_parameter class, which stores the value of a single parameter in tao.
Also defined are a few functions for parsing data into a single tao_parameter or a dictionary of tao_parameters
'''
from collections import OrderedDict
import string

tao_startup_param_list = [
    'beam_file;FILE;T;',
    'beam_init_position_file;FILE;T;',
    'beam_track_data_file;FILE;T;',
    'building_wall_file;FILE;T;',
    'data_file;FILE;T;',
    'hook_init_file;FILE;T;',
    'init_file;FILE;T;',
    'noinit;LOGIC;T;F',
    'lattice_file;FILE;T;',
    'plot_file;FILE;T;',
    'startup_file;FILE;T;',
    'noplot;LOGIC;T;F',
    'var_file;FILE;T;',
    'slice_lattice;FILE;T;',
    'disable_smooth_line_calc;LOGIC;T;F',
    'log_startup;LOGIC;T;F',
    'no_stopping;LOGIC;T;F',
    'rf_on;LOGIC;T;F',
]

#-------------------------------------------------

class tao_parameter():
    '''
    Basic class for holding the properties of a parameter in Tao.

    name:           The name of the parameter
    type:           "STR", "INT", "REAL", "LOGIC", "ENUM", etc...
    can_vary:       Either 'T', 'F', or 'I', indicating whether or not the
                    user may change the value of the paramter. 'I' indicates
                    that the parameter is to be ignored by the gui, except
                    to possibly be displayed as a sub-parameter to another
                    tao_parameter.
    value:          The value held in the parameter, should be of the
                    appropriate type for the specified param_type
                    (or 'T'/'F' for LOGIC)
    NOTE: It is unclear if sub_param is actually ever set by a Tao python command. -- DCS 11/2020
    sub_param:      The name of the sub_parameter associated with this parameter,
                    For example: ele_name has the sub parameter ix_ele.
    '''

    def __init__(self, param_name, param_type, can_vary, param_value, sub_param=None):
        # Enums and inums may have a prefix attached to their name, as in
        # axis^type.  In this case, the prefix is removed from the parameter name
        # and stored in the self.prefix variable
        if (param_type in ['ENUM', 'INUM']) and (param_name.count('^') == 1):
            self.prefix, self.name = param_name.split('^')
        else:
            self.prefix = None
            self.name = param_name
        self.type = param_type
        self.can_vary = (can_vary == 'T')
        self.is_ignored = (can_vary == 'I')
        self.sub_param = sub_param #associated sub_parameter (name)

        if param_type in ['STR', 'FILE', 'DAT_TYPE', 'DAT_TYPE_Z', 'DAT_TYPE_E',
                'REAL_ARR', 'ENUM', 'ENUM_Z', 'STRUCT', 'COMPONENT', 'SPECIES']:
            self.value = param_value
        elif param_type == 'INT':
            try:
                self.value = int(param_value)
            except:
                self.value = None
        elif param_type == 'REAL':
            try:
                self.value = float(param_value)
            except:
                self.value = None
        elif param_type == 'LOGIC':
            self.value = (param_value == 'T')
        elif param_type == 'INUM':
            try:
                self.value = int(param_value)
            except:
                self.value = None
        else:
            print ('UNKNOWN PARAMETER TYPE: ' + param_type)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return self.name + ';' + self.type + ';' + str(self.can_vary) + ';' + str(self.value)

    def get_component(self, comp_name):
        '''
        Looks for a component called comp_name in self, and returns the component
        with the matching name
        Returns None if comp_name was not found or if self.type != STRUCT
        '''
        if self.type != 'STRUCT':
            return None
        for param in self.value:
            if param.name == comp_name:
                return param.value
        return None

#

class InvalidParamError(Exception):
    '''
    Provides an exception for when a param_string is improperly formatted
    Examples of improper formatting include: not enough semicolons
    (3 minimum), bad sub-parameter formatting (for STRUCTs), etc
    '''
    pass

# An item in the parameter list is a string that looks like:

def tao_parameter_dict(param_list):
    '''
    Takes a list of strings, each string looks something like: 'param_name;STR;T;abcd'
    and returns a dictionary with keys being the param_name.
    Blank strings will be ignored.
    '''
    this_dict = OrderedDict()
    for param_str in param_list:
        if param_str.strip() == '': continue
        v = param_str.split(';')
        this_dict[v[0]] = str_to_tao_param(param_str)
    return this_dict

#

def str_to_tao_param(param_str):
    '''
    Takes a parameter string (EG: 'param_name;STR;T;abcd')
    and returns a tao_parameter instance
    param_str MUST have at least 3 semicolons
    '''
    v = param_str.split(';')
    if len(v) < 3:
        msg = str(param_str) + " is not a valid param_string (not enough semicolons)"
        raise InvalidParamError(msg)
    sub_param = None #default
    #TEMPORARY FIX
    if (len(v[2]) == 2) & (len(v) == 3):
        v.append(v[2][1])
        v[2] = v[2][0]
    ###
    # Special case: REAL_ARR (unknown length)
    if v[1] == "REAL_ARR":
        arr = []
        for i in range(len(v[3:])):
            x = v[3:][i]
            try:
                arr.append(float(x))
            except:
                if i==len(v[3:])-1: #last item, could be a related parameter name
                    if len(x) > 0:
                        sub_param = x
                else:
                    arr.append(float(0))
        v[3] = arr
    elif v[1] == 'STRUCT':
        n_comp = int(len(v[3:])/3)
        components = v[3:][:3*n_comp]
        c_list = [0]*n_comp
        for n in range(n_comp):
            c_name = components[3*n]
            c_type = components[3*n+1]
            c_val = components[3*n+2]
            c_list[n] = tao_parameter(c_name, c_type, v[2], c_val)
        v[3] = c_list
        if len(v[3:]) % 3 == 1: # one more item than expected --> it is a sub_param
            sub_param = v[-1]
    # Generic case sub_param: name;type;can_vary;value;sub_param
    if len(v) == 5:
        sub_param=v[4]
    return tao_parameter(v[0],v[1],v[2],v[3], sub_param)

#-------------------------------------------------
tao_startup_param_dict = tao_parameter_dict(tao_startup_param_list)



'''
This module defines the tao_interface class, a general purpose interface
between tao and python that supports both the ctypes and pexpect
backend implementations.  Also defined here are a couple of helper functions
for the tao_interface class.
'''
import sys
import os
import re
import io

from .tao_pexpect import tao_io
from .tao_ctypes import Tao

class new_stdout(object):
    '''
    Re-routes the print statements generated by tao_io
    so that they can be shown by tao_console.
    '''
    def __init__(self):
        self.file_obj = io.StringIO()
    def __enter__(self):
        self.current_out = sys.stdout
        sys.stdout = self.file_obj
        return self.file_obj
    def __exit__(self, type, value, traceback):
        sys.stdout = self.current_out

def filter_output(x):
    '''
    Filters out certain ANSI escape sequences from the string x
    '''
    if not isinstance(x, str):
        return x
    # Filter out \[[6 q first
    x = x.replace('\x1b[6 q', '')
    # replace line feed ('\x0a') with return ('\x0d')
    #while x.find('\x0a') != -1:
    #  x = x.replace('\x0a', '\x0d')
    # Remove xterm-specific escape code
    if x.find('\x1b[?1034h') != -1:
        x = x.replace('\x1b[?1034h', "")
    # Filter out color codes
    color_regex = re.compile('\x1b\\[[0-9;]*m')
    matches = color_regex.findall(x)
    for color in matches:
        x = x.replace(color, '')
    return x


class tao_interface():
    '''
    Serves as a general purpose interface between tao and python
    Supports both the ctypes and pexpect interface implementations, allowing
    the user to select between them and falling back on the other if the
    selected interface cannot be used

    Init arguments:
    mode: either "ctypes" (default) or "pexpect"
    init_args: a string with the command line initialization arguments that
            tao should use, e.g. "-init tao.init -rf_on
    tao_exe: the tao executable to use (applies to pexpect only)
    so_lib: the tao shared library to use (applies to ctypes only)
    expect_str: the prompt string that tao will use (default is "Tao>")

    If you need direct access to the methods defined by the ctypes or pexpect
    backends, you can call them on self.pexpect_pipe and self.ctypes_pipe
    respectively, e.g. in ctypes mode, the cmd_real method can be accessed by
    self.ctypes_pipe.cmd_real()
    '''
    def __init__(self, mode="ctypes", init_args = "", tao_exe =  "", expect_str = "Tao>", so_lib=""):
        self.mode = mode
        self.exe_lib_warnings = ""
        self.exe_lib_warning_type = 'normal'
        # Where to look for the shared library/executable
        if 'ACC_LOCAL_ROOT' in os.environ.keys():
            EXE_DIR = os.environ['ACC_LOCAL_ROOT'] + '/production/bin/'
            LIB_DIR = os.environ['ACC_LOCAL_ROOT'] + '/production/lib/'
        else:
            EXE_DIR = os.environ['ACC_EXE'] + '/'
            LIB_DIR = None
        # Check for executable
        if os.path.isfile(tao_exe) and os.access(tao_exe, os.X_OK):
            exe_found = True
        elif os.path.isfile(EXE_DIR + 'tao') and os.access(EXE_DIR + 'tao', os.X_OK):
            tao_exe = EXE_DIR + 'tao'
            exe_found = True
            if (mode == "pexpect") & (tao_exe != ""):
                self.exe_lib_warnings += "Note: could not find " + tao_exe
                self.exe_lib_warnings += ".\nDefaulting to " + EXE_DIR + 'tao'
        else:
            exe_found = False
            if mode == "pexpect":
                self.exe_lib_warnings += "Warning: no executable found, defaulting to ctypes"
                self.exe_lib_warning_type = 'error'

        # Check for shared library (and set up self.ctypes_pipe)
        lib_found = False
        try:
            self.ctypes_pipe = Tao(so_lib=so_lib)
            lib_found = True
        except OSError: #so_lib not found
            if LIB_DIR != None:
                try:
                    self.ctypes_pipe = Tao(so_lib=LIB_DIR+'libtao.so')
                    lib_found = True
                    if mode == "ctypes":
                        self.exe_lib_warnings += "Note: could not find " + so_lib
                        self.exe_lib_warnings += ".\nUsing library in " + LIB_DIR
                except OSError: #so_lib not found
                    pass # will continue below
            if not lib_found:
                try:
                    self.ctypes_pipe = Tao(so_lib="")
                    lib_found = True
                    if mode == "ctypes":
                        self.exe_lib_warnings += "Note: could not find " + so_lib
                        self.exe_lib_warnings += ".\nUsing library in " + os.environ['ACC_ROOT_DIR']+'/production/lib/'
                except ValueError:
                    lib_found = False
                    if mode == "ctypes":
                        self.exe_lib_warnings += "Warning: no shared library found, defaulting to pexpect"
                        self.exe_lib_warning_type = 'error'
        except ValueError:
            lib_found = False

        if self.exe_lib_warnings != "":
            self.exe_lib_warnings += '\n'

        # Switch mode if necessary
        if exe_found and not lib_found:
            mode = 'pexpect'
            self.mode = 'pexpect'
        elif not exe_found and lib_found:
            mode = 'ctypes'
            self.mode = 'ctypes'
        elif not exe_found and not lib_found:
            mode = 'failed'
            self.mode='failed'

        # new_stdout() needed to capture print statements from tao_io
        with new_stdout() as output:
            if mode == "pexpect":
                self.pexpect_pipe = tao_io(init_args=init_args,
                        tao_exe=tao_exe, expect_str=expect_str)
            if mode == "ctypes":
                for line in self.ctypes_pipe.init(init_args):
                    print(line)
            if mode == 'failed':
                self.exe_lib_warnings += "FATAL: could not locate Tao shared library or executable.\n"
                self.exe_lib_warnings += "Please reinitialize with a good executable or shared library."
                self.exe_lib_warning_type = "fatal"
                print("")

        # Process startup message
        output = output.getvalue()
        output = filter_output(output)
        self.startup_message = output

    def cmd_in(self, cmd_str):
        '''
        Runs cmd_str at the Tao command line and returns the output as a string
        '''
        # Run the command:
        if self.mode=="pexpect":
            output = self.pexpect_pipe.cmd_in(cmd_str)
        elif self.mode=='ctypes':
            output_list = self.ctypes_pipe.cmd(cmd_str)
            output = ""
            for line in output_list:
                if line.strip != '':
                    output += line + '\n'
            if (len(output) > 0) and (output[-1] == '\n'):
                output = output[:-1]
        else:
            output = ""

        # Scrub output for extra new lines at the end,
        # as well as escape characters
        if cmd_str.find('python') == 0:
            output = output.replace('\r\n\r\n', '')
        output = filter_output(output)

        return output

    def cmd(self, cmd_str):
        '''
        Runs cmd_str at the Tao command line and prints the output
        to standard out
        '''
        output = self.cmd_in(cmd_str)



'''
Contains the base classes used by many (if not all) of the windows
in the GUI for tao.
'''
import sys
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
import copy
from .tao_widget import *
from .tao_set import *

#-----------------------------------------------------
# Tao_Toplevel window
class Tao_Toplevel(tk.Toplevel):
    '''
    Skeleton class that handles adding/removing windows
    from the root refresh windows dictionary
    Subclasses should set self.tao_id before running this
    __init__ method.
    Subclasses MUST set self.root before running this
    __init__ method.
    '''
    def __init__(self, parent, *args, **kwargs):
        tk.Toplevel.__init__(self, parent, class_='Tao', *args, **kwargs)
        if sys.platform == "linux":
            self.iconbitmap(self.root.icon)
        # Handle root window list placement
        if 'tao_id' not in self.__dict__:
            self.tao_id = None
        for tao_id in self.root.tao_id_list:
            if self.tao_id == tao_id:
                self.root.refresh_windows[tao_id].append(self)
                break

    def destroy(self):
        '''
        Overloaded here to remove self from root.refresh_windows
        '''
        # Try to remove self from root.refresh_windows
        for tao_id in self.root.tao_id_list:
            if self.tao_id == tao_id:
                try:
                    self.root.refresh_windows[tao_id].pop(self.root.refresh_windows[tao_id].index(self))
                except ValueError:
                    pass
        # Call Toplevel.destroy()
        tk.Toplevel.destroy(self)


class Tao_Popup(Tao_Toplevel):
    '''
    Provides a more convenient way to initialize a toplevel
    window that doesn't need to be its own subclass, used for
    things like ele_shape windows, dialog boxes, etc

    parent: the parent window
    root: the tao root window
    tao_id: which set of refresh windows this window should belong to (optional)
    '''
    def __init__(self, parent, root, tao_id=None, *args, **kwargs):
        self.root = root
        self.tao_id = tao_id
        Tao_Toplevel.__init__(self, parent, *args, **kwargs)


#-----------------------------------------------------
# List window

class tao_list_window(Tao_Toplevel):
    '''
    Skeleton class with several commonly used features.
    Comes with a scrollable region and supports adding other widgets
    Subclasses set tao_id before running this __init__ to handle
    adding/removing themselves from root window lists
    '''
    def __init__(self, root, title, use_upper=False,
            min_width=0, parent=None, *args, **kwargs):
        self.root = root
        if parent != None:
            Tao_Toplevel.__init__(self, parent, *args, **kwargs)
        else:
            Tao_Toplevel.__init__(self, root, *args, **kwargs)
        self.title(title)

        #self.wm_geometry(newGeometry='400x600')

        self.upper_frame=tk.Frame(self)
        self.outer_frame=tk.Frame(self)

        self.canvas=tk.Canvas(self.outer_frame)
        self.list_frame=tk.Frame(self.canvas)
        scrollbar=tk.Scrollbar(self.outer_frame,orient="vertical",
                command=self.canvas.yview)
        self.canvas.configure(yscrollcommand=scrollbar.set)

        def scrollhelper(event):
            self.canvas.configure(scrollregion=self.canvas.bbox("all"))
            new_width = event.width + 15
            # Don't resize to a smaller size
            old_geo = self.wm_geometry()
            old_width = int(old_geo[:old_geo.find('x')])
            old_height = int(old_geo[old_geo.find('x')+1:old_geo.find('+')])
            if old_width > new_width:
                new_width = old_width
            # Don't resize to be smaller than the
            # upper frame
            self.upper_frame.update()
            upper_width = self.upper_frame.winfo_width()
            if upper_width > new_width:
                new_width = upper_width
            # Don't resize to less than the min width
            if min_width > new_width:
                new_width = min_width
            if old_height != 1:
                #new_geo = old_geo[old_geo.find('x'):old_geo.find('+')]
                new_geo = old_geo[old_geo.find('x'):]
                new_geo = str(new_width) + new_geo
            else:
                new_geo = str(new_width) + 'x500'
            # Place the new window near the root window
            if new_geo.find('+') == -1:
                new_pos = self.root.wm_geometry()
                new_pos = new_pos[new_pos.find('+'):]
                new_geo = new_geo + new_pos
            self.geometry(new_geo)
            self.canvas.configure(width=new_width)
        self.list_frame.bind("<Configure>",scrollhelper)

        self.outer_frame.bind("<Enter>", self.bind_mouse)
        self.outer_frame.bind("<Leave>", self.unbind_mouse)

        if use_upper:
            self.upper_frame.pack(side="top",fill="both", expand=0)
        self.outer_frame.pack(side="top",fill="both",expand=1)
        scrollbar.pack(side="right",fill="y")
        self.canvas.pack(side="left",fill="both",expand=1)

        self.canvas_list_window = self.canvas.create_window(
                (0,0),window=self.list_frame,anchor='nw')

    def bind_mouse(self, event):
        self.outer_frame.bind_all("<Button-4>", self.mouse_scroll)
        self.outer_frame.bind_all("<Button-5>", self.mouse_scroll)

    def unbind_mouse(self, event):
        self.outer_frame.unbind_all("<Button-4>")
        self.outer_frame.unbind_all("<Button-5>")

    def mouse_scroll(self, event):
        #self.canvas.yview_scroll(direction,"units")
        try:
            if event.num == 4:
                self.canvas.yview_scroll(-1,"units")
            elif event.num == 5:
                self.canvas.yview_scroll(1,"units")
        except: #in case this gets called when it shouldn't
            pass

#-----------------------------------------------------
# Scrolling frame
class tao_scroll_frame(tk.Frame):
    '''
    Provides a frame with a vertical scrollbar attached
    parent: the parent of this frame
    min_width: the minimum width that this frame must have
    self.frame should be used for placing widgets into
    this frame
    '''
    #TODO: upgrade tao_list_window to use this for its scrolling
    def __init__(self, parent, min_width=0, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.min_width = min_width
        self._old_width=0
        self._canvas=tk.Canvas(self)
        self.frame=tk.Frame(self._canvas)
        scrollbar=tk.Scrollbar(self,orient="vertical",
                command=self._canvas.yview)
        self._canvas.configure(yscrollcommand=scrollbar.set)

        self.frame.bind("<Configure>", self.frame_handler)
        self._canvas.bind("<Configure>", self.canvas_handler)

        self.bind("<Enter>", self.bind_mouse)
        self.bind("<Leave>", self.unbind_mouse)

        self._canvas.grid(row=0, column=0, sticky='NSEW')
        scrollbar.grid(row=0, column=1, sticky='NS')
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self._canvas_window = self._canvas.create_window(
                (0,0),window=self.frame,anchor='nw')

    def frame_handler(self, event):
        self._canvas.configure(scrollregion=self._canvas.bbox("all"))
        frame_width = self.frame.winfo_reqwidth()
        #frame_width = event.width
        # try using winfo_reqwidth() TODO
        self.grid_columnconfigure(0, minsize=frame_width)
        self.update_idletasks()

    def canvas_handler(self, event):
        #canvas_width = self._canvas.winfo_reqwidth()#event.width
        canvas_width = event.width
        self._canvas.itemconfig(self._canvas_window, width=canvas_width)
        self.update_idletasks()

    def scrollhelper(self, event):
        self._canvas.configure(scrollregion=self._canvas.bbox("all"))
        new_width = self.frame.winfo_geometry()
        new_width = int(new_width[:new_width.find('x')])
        # Don't resize to a smaller size
        #if self._old_width > new_width:
        #    return
        # Don't resize to less than the min width
        if self.min_width > new_width:
            new_width = self.min_width
        self.grid_columnconfigure(0, minsize=new_width)
        #self._canvas.configure(width=new_width)
        # Set the width of the frame to match the canvas
        new_frame_width = self._canvas.winfo_width()
        if new_frame_width > 1:
            self._canvas.itemconfigure("self.frame", width=new_frame_width)
        self._old_width = new_width

    def bind_mouse(self, event):
        self.bind_all("<Button-4>", self.mouse_scroll)
        self.bind_all("<Button-5>", self.mouse_scroll)

    def unbind_mouse(self, event):
        self.unbind_all("<Button-4>")
        self.unbind_all("<Button-5>")

    def mouse_scroll(self, event):
        #self.canvas.yview_scroll(direction,"units")
        try:
            if event.num == 4:
                self._canvas.yview_scroll(-1,"units")
            elif event.num == 5:
                self._canvas.yview_scroll(1,"units")
        except: #in case this gets called when it shouldn't
            pass

#-----------------------------------------------------
# Parameter frame
class tao_parameter_frame(tk.Frame):
    '''
    Meant to display a list of parameters in a given
    number of columns
    tao_list should be a list of tao_parameters
    '''
    def __init__(self, parent, tao_list, n_col, pipe, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.pipe = pipe
        self.tao_list = [] #List for tk_tao_parameters
        for p in tao_list:
            self.tao_list.append(tk_tao_parameter(p, self, pipe))
        cols = [] # A list of the parts of tao_list,
        # after being divided into n_col columns
        # Don't count units# in list length
        real_len = len(self.tao_list)
        for item in self.tao_list:
            if item.param.name.find('units#') == 0:
                real_len -= 1
        # real_len now actually represents how many items there are
        len_col = int(real_len/n_col) # How many elements per column
        if len_col < len(self.tao_list)/n_col:
            #if the result was rounded down
            len_col = len_col+1

        # Grid widgets (and units)
        i = 0
        for item in self.tao_list:
            if item.param.name.find('units#') != 0:
                r = (i % len_col) + 2 # rows 0,1 reserved
                c = int(i / len_col) * 3
                item.tk_label.grid(row=r, column=c, sticky='E')
                item.tk_wid.grid(row=r, column=c+1, sticky='EW')
                i = i+1
            else:
                tk.Label(self,text=item.param.value).grid(row=r, column=c+2, sticky='W')

    def set_params(self, set_str, event=None):
        '''
        Runs tao_set on self.tao_list
        '''
        tao_set(self.tao_list, set_str, self.pipe)

    def check_for_changes(self):
        '''
        Runs check_for_changes (from tao_set) on self.tao_list
        '''
        return check_for_changes(self.tao_list)


#-----------------------------------------------------
# Parameter window
class tao_parameter_window(tao_list_window):
    '''
    Generic window for displaying and editing a list of parameters
    '''
    def __init__(self, root, title, tao_list, pipe, plot="", parent=None, *args, **kwargs):
        tao_list_window.__init__(self, root, title, parent=parent, *args, **kwargs)
        self.button_frame = tk.Frame(self)
        self.button_frame.pack(side="top", fill="both", expand=0)
        self.tao_list = tao_list
        for k in range(len(self.tao_list)):
            self.tao_list[k] = tk_tao_parameter(
                    self.tao_list[k], self.list_frame, pipe, plot=plot)
        # Link/filter ignored parameters
        self.tao_list = tk_tao_linker(self.tao_list)
        for k in range(len(self.tao_list)):
            self.tao_list[k].tk_label.grid(row=k,column=0,sticky="W")
            self.tao_list[k].tk_wid.grid(row=k,column=1,sticky="EW")
            if self.tao_list[k].sub_wid != None:
                self.tao_list[k].sub_wid.grid(row=k, column=2, sticky='W')
        # Attempt to bind data_source and data_types together
        data_source_ix = None
        for k in range(len(self.tao_list)):
            if self.tao_list[k].param.name == "data_source":
                data_source_ix = k
                break
        if data_source_ix != None:
            # Define _data_source_handler
            def _dat_source_handler(*args):
                '''Updates _data_source for DAT_TYPE widgets in this window'''
                for w in self.tao_list:
                    if w.param.type == 'DAT_TYPE':
                        w._data_source = self.tao_list[data_source_ix].tk_var.get()
                        w._s_refresh(dat_source_swap=True)
            # Trace tk_var for the data_source widget
            self.tao_list[data_source_ix].tk_var.trace('w', _dat_source_handler)
            _dat_source_handler()

#-----------------------------------------------------
# Table window
class table_window(tao_list_window):
    '''
    Meant for showing large amounts of information in a table
    (e.g. d1_data and v1_vars).  Comes with bulk editing, detailed view of
    individual rows, and editing of individual parameters in the table.

    Input parameters:
    root: parent widget
    pipe: the tao_interface object that allows interface with Tao
    array_name: the name of the object this table displays
            (e.g. the name of the d1_datum or v1_variable)
    title_list: the column titles, in order from left to right
    bulk_template: a list with elements [tao_parameter, column], where
            tao_parameter is a generic copy of the parameter that will be bulk filled
            (i.e. initialized to "blank"), and column is the column number it should
            be gridded to (start counting from 0 on the left)
    bulk_set_format: format string for the bulk set string, to be used with the
            str.format() method
    set_format: format string for individual rows' set strings, to be used with
            the str.format() method
            example format strings: "set data 1@{}|", "set data 1@{}[{}]|"
    '''

    def __init__(self, root, pipe, array_name, title_list, bulk_template,
            bulk_set_format, set_format, *args, **kwargs):
        tao_list_window.__init__(self, root, array_name, *args, **kwargs)
        self.pipe = pipe
        self.array_name = array_name
        self.title_list = title_list
        self.bulk_template = bulk_template
        self.bulk_set_format = bulk_set_format
        self.set_format = set_format
        self.refresh()

        self.button_frame = tk.Frame(self)
        self.button_frame.pack(side="bottom", fill="both", expand=0)
        b1 = tk.Button(self.button_frame, text="Apply Changes", command=self.apply)
        b2 = tk.Button(self.button_frame, text="Discard Changes and Refresh",
                command=self.refresh)

        b2.pack(side="right")
        b1.pack(side="right")

    def refresh(self):
        # Clear self.list_frame:
        for child in self.list_frame.winfo_children():
            child.destroy()

        # Grid the column titles
        j = 0
        for item in self.title_list:
            tk.Label(self.list_frame, text=item).grid(row=0, column=j)
            self.list_frame.grid_columnconfigure(j, pad=10)
            j=j+1

        # Bulk editing
        tk.Label(self.list_frame, text="Bulk editing:").grid(
                row=1, column=0, columnspan=self.bulk_template[0][1])
        tk.Label(self.list_frame, text="Click to fill:").grid(
                row=2, column=0, columnspan=self.bulk_template[0][1])
        #self.bulk_template[0][1] is the appropriate
        #columnspan because these labels can fill all
        #the space until the first bulk fill box

        self.bulk_params = [] #Holds the current bulk edit widgets
        self.bulk_filled = [] #Holds whether or not bulk filling has been used
        self.bulk_value = [] #Holds the last value that was filled to the cells
        self.bulk_apply = [] #Holds fill buttons

        j=0
        for item in self.bulk_template:
            self.bulk_params.append(tk_tao_parameter(copy.copy(item[0]),
                self.list_frame, self.pipe))
            self.bulk_params[j].tk_wid.grid(row=1, column=item[1])
            self.bulk_params[j].tk_wid.bind("<Return>", self.fill_callback(j))
            self.bulk_filled.append(False)
            self.bulk_value.append(self.bulk_params[j].tk_var.get())
            self.bulk_apply.append(tk.Button(self.list_frame, text="Fill...",
                command = self.fill_callback(j) ))
            self.bulk_apply[j].grid(row=2, column=item[1])
            j=j+1

        #Fetch and fill in the data
        # IT IS EXPECTED THAT SUBCLASSES WILL DEFINE
        # self.row_num AND A METHOD FOR POPULATING
        # self.list_rows, AND THEN CALL THIS METHOD
        #i = row counter, j = column counter
        #grid to row i+3 because row 0 is titles, row 1 is bulk editing widgets,
        #row 2 is fill buttons
        self.list_rows = []
        for i in range(self.row_num):
            self.list_rows.append(self.list_row_fetch(i))
            for j in range(len(self.list_rows[i].tk_wids)):
                self.list_rows[i].tk_wids[j].grid(row=i+3, column=j)
            tk.Button(self.list_frame, text="View More...",
                    command=self.open_detail_window_callback(
                        self.list_rows[i].index)).grid(
                                row=i+3, column=len(self.list_rows[i].tk_wids))

    def fill(self, index, event=None):
        '''
        Fills the column specified by bulk_template to the bulk edit value,
        saves the bulk edit value for efficient calls to tao_set,
        and clears the bulk edit box
        '''
        # Save the bulk parameter state
        self.bulk_value[index] = self.bulk_params[index].tk_var.get()
        self.bulk_filled[index] = True
        # Clear the bulk parameter widget
        if self.bulk_params[index].param.type in ['STR', 'INT', 'REAL']:
            self.bulk_params[index].tk_var.set("")
        elif self.bulk_params[index].param.type == 'LOGIC':
            self.bulk_params[index].tk_var.set(False)
        # Fill the appropriate variable
        for i in range(len(self.list_rows)):
            self.list_rows[i].tk_tao_params[
                    self.bulk_template[index][0].name].tk_var.set(self.bulk_value[index])

    def fill_callback(self, index, event=None):
        return lambda event=None : self.fill(index)


    def apply(self):
        #Apply bulk changes
        for i in range(len(self.bulk_params)):
            if self.bulk_filled[i]:
                set_str = self.bulk_set_format.format(self.array_name)
                self.bulk_params[i].tk_var.set(self.bulk_value[i])
                #overide is necessary for LOGIC parameters
                tao_set([self.bulk_params[i]], set_str, self.pipe, overide=(
                    self.bulk_params[i].param.type=='LOGIC'))

        #Apply individual changes that are different from bulk changes
        for i in range(len(self.list_rows)):
            set_list = []
            #NOTE: it is expected that self.list_rows[i].index exist,
            #and be equal to that row's index
            set_str = self.set_format.format(self.array_name, self.list_rows[i].index)

            #Find elements in row that need setting
            for j in range(len(self.bulk_template)):
                name = self.bulk_template[j][0].name
                c1 = (self.list_rows[i].tk_tao_params[name].tk_var.get()
                        != self.bulk_value[j])
                c2 = not self.bulk_filled[j]
                try:
                    if self.bulk_template[j][0].type == 'REAL':
                        c3 = (float(self.list_rows[i].tk_tao_params[name].tk_var.get())
                                != self.list_rows[i].tk_tao_params[name].param.value)
                    elif self.bulk_template[j][0].type == 'INT':
                        c3 = (int(self.list_rows[i].tk_tao_params[name].tk_var.get())
                                != self.list_rows[i].tk_tao_params[name].param.value)
                    elif self.bulk_template[j][0].type == 'STR':
                        c3 = (str(self.list_rows[i].tk_tao_params[name].tk_var.get())
                                != self.list_rows[i].tk_tao_params[name].param.value)
                    elif self.bulk_template[j][0].type == 'LOGIC':
                        c3 = (bool(self.list_rows[i].tk_tao_params[name].tk_var.get())
                                != self.list_rows[i].tk_tao_params[name].param.value)
                except ValueError:
                    c3 = False
                if (c1 | c2) & c3:
                    set_list.append(self.list_rows[i].tk_tao_params[name])

            if set_list != []:
                tao_set(set_list, set_str, self.pipe)

        #Refresh
        self.refresh()

    def open_detail_window_callback(self, index):
        return lambda : self.open_detail_window(index)

    def open_detail_window(self, index):
        # SUBCLASSES ARE EXPECTED TO DEFINE self.param_list
        # FOR USE IN A tao_parameter_window, THEN CALL
        # THIS METHOD
        detail_title = self.array_name + '[' + str(index) + ']'
        win = tao_parameter_window(self.root, detail_title, self.param_list, self.pipe, parent=self)

        set_str = self.set_format.format(self.array_name, index)
        b = tk.Button(win.button_frame, text="Apply changes",
                command=lambda : self.detail_set_callback(win.tao_list,set_str))
        b.pack()

    def detail_set_callback(self, tao_list, set_str):
        tao_set(tao_list, set_str, self.pipe)
        self.refresh()

#-----------------------------------------------------
class lw_table_window(Tao_Toplevel):
    '''
    Light-weight version of table_window, intended to be less
    graphically/computationally intensive (for use with networked
    machines, for example).  Table is read only but individual
    entries can still be opened for editing with double click.
    '''
    def __init__(self, root, pipe, array_name, title_list, bulk_format,
            bulk_set_format, set_format, *args, **kwargs):
        self.root = root
        Tao_Toplevel.__init__(self, root, *args, **kwargs)
        self.title(array_name)
        self.pipe = pipe
        self.array_name = array_name
        self.title_list = title_list
        self.bulk_format = bulk_format
        self.bulk_set_format = bulk_set_format
        self.set_format = set_format
        self.button_frame = tk.Frame(self) #holds the buttons
        self.button_frame.pack(fill='x', expand=0)
        self.table_frame = tk.Frame(self) #holds the table
        self.table_frame.pack(fill='both', expand=1)

        if len(bulk_format) > 0:
            tk.Button(self.button_frame, text="Bulk fill...",
                    command=self.open_bulk_window).pack(side='left')

        self.sh_button = tk.Button(self.button_frame, text="Show all",
                command=self.toggle_show_all)
        self.sh_button.pack(side='left')
        self.hide_nonexist = True
        self.refresh()

    def toggle_show_all(self, event=None):
        self.hide_nonexist = not self.hide_nonexist
        newtext = ("Show all" if self.hide_nonexist
                else "Hide nonexistent rows")
        self.sh_button.configure(text=newtext)
        self.refresh()

    def refresh(self):
        '''
        Creates a treeview to display the table information,
        and binds double click to open one item
        '''
        # Clear the existing table
        for child in self.table_frame.winfo_children():
            child.destroy()

        widths = [0]*len(self.title_list) # tracks column widths

        # Create table
        self.tree = ttk.Treeview(
                self.table_frame, columns=self.title_list, show='headings')
        # Column titles
        for title in self.title_list:
            self.tree.heading(title, text=title)
            self.tree.column(title, stretch=True, anchor='center')

        # Fill rows
        # Fetch and fill in the data
        # IT IS EXPECTED THAT SUBCLASSES WILL DEFINE
        # self.row_num AND A METHOD FOR POPULATING
        # self.list_rows, AND THEN CALL THIS METHOD
        #i = row counter, j = column counter
        self.list_rows = []
        for i in range(self.row_num):
            row = self.lw_list_row_fetch(i)
            if self.hide_nonexist:
                # Filter out if exists is False
                if row[-1] == 'F':
                    continue
            self.tree.insert("", "end", values=row)
            for j in range(len(row)):
                if len(row[j])*12 > widths[j]:
                    widths[j] = len(row[j])*12

        # Set column widths appropriately
        for j in range(len(self.title_list)):
            if len(self.title_list[j])*12 > widths[j]:
                widths[j] = len(self.title_list[j])*12
            self.tree.column(self.title_list[j], width=widths[j], minwidth=widths[j])

        # Scrollbars
        hbar = ttk.Scrollbar(
                self.table_frame, orient="horizontal", command=self.tree.xview)
        vbar = ttk.Scrollbar(
                self.table_frame, orient="vertical", command=self.tree.yview)
        self.tree.configure(xscrollcommand=hbar.set, yscrollcommand=vbar.set)

        vbar.pack(side="right", fill="y", expand=0)
        hbar.pack(side="bottom", fill='x', expand=0)
        self.tree.pack(side="left", fill="both", expand=1)
        self.widths = widths

        # Double click to open element
        self.tree.bind('<Double-Button-1>', self.lw_detail_callback)

        tot = 0
        for w in widths:
            tot = tot+w
        self.maxsize(1800, 1000)
        self.minsize(1300, 100)

    def open_detail_window_callback(self, event=None):
        '''
        Checks the table for the currently selected row and
        makes a call to open_detail_window_callback
        '''
        x = self.tree.focus()
        row = self.tree.item(x)
        self.open_detail_window(int(row['values'][0]))

    def open_bulk_window(self, event=None):
        '''
        Opens a window with bulk settings for meas_value,
        ref_value, weight, and good_user.
        '''
        win = Tao_Toplevel.__new__(Tao_Toplevel)
        win.root = self.root
        Tao_Toplevel.__init__(win, self)
        win.title("Bulk settings for " + self.array_name)

        j = 0 #column counter
        fill_choices = [] #what is being used to fill (tk.StringVar()'s)
        for item in self.bulk_format:
            fill_choices.append(j)
            fill_frame, where_frame, fill_choices[j] = self.make_bulk_frame(
                    win, item[0], usebmd=(item[0].name=='meas_value'))
            fill_frame.grid(row=0, column=2*j, sticky='NSEW')
            where_frame.grid(row=2, column=2*j, sticky='NSEW')
            if j != 0:
                ttk.Separator(win, orient='vertical').grid(
                        row=0, column=2*j-1, rowspan=3, sticky='NS')
            win.columnconfigure(j, weight=1)
            j = j+1
        ttk.Separator(win, orient='horizontal').grid(
                row=1, column=0, columnspan=2*j-1, sticky='EW')
        ttk.Separator(win, orient='horizontal').grid(
                row=3, column=0, columnspan=2*j-1, sticky='EW')

        def fill_cmd(event=None):
            self.bulk_set(fill_choices, win)

        tk.Button(win, text="Fill and apply", command=fill_cmd).grid(
                row=4, column=0, columnspan=2*j-1)

    def bulk_set(self, fill_choices, parent):
        '''
        Runs set commands to set variables specified in a bulk window
        appropriately.
        '''
        for i in range(len(self.bulk_format)):
            if fill_choices[i]['choice'].get() == 'none':
                continue
            else:
                fill_val = fill_choices[i][fill_choices[i]['choice'].get()].get()

            # Check that the formula gives the correct length array
            if fill_choices[i]['choice'].get() == 'formula':
                if fill_choices[i]['where'].get() == 'all':
                    len1 = len(self.pipe.cmd_in('python evaluate '
                        + self.array_name + '|'
                        + self.bulk_format[i][0].name).splitlines())
                elif fill_choices[i]['where'].get() == 'range':
                    len1 = len(self.pipe.cmd_in('python evaluate '
                        + self.array_name + '[' + fill_choices[i]['range'].get()
                        + ']|' + self.bulk_format[i][0].name).splitlines())
                len2 = len(self.pipe.cmd_in('python evaluate '
                    + fill_val).splitlines())
                if len1 != len2:
                    messagebox.showwarning('Error',
                            'Entered formula yields an array of incorrect length.',
                            parent=parent)
                    continue

            if fill_choices[i]['where'].get() == 'all':
                set_str = self.bulk_set_format.format(self.array_name)
            elif fill_choices[i]['where'].get() == 'range':
                set_str = self.bulk_set_format.format(self.array_name + '['
                        + fill_choices[i]['range'].get() + ']')

            if fill_choices[i]['choice'].get() == 'const':
                # Make sure fill_val is True/False for LOGIC parameters
                if self.bulk_format[i][0].type == 'LOGIC':
                    fill_val = str(bool(fill_val))
                self.pipe.cmd_in(set_str + self.bulk_format[i][0].name
                        + ' = ' + fill_val)
            elif fill_choices[i]['choice'].get() == 'bmd':
                fill_val = fill_val.lower()
                self.pipe.cmd_in(set_str + self.bulk_format[i][0].name
                        + ' = ' + set_str.split(' ')[-1] + fill_val)
            elif fill_choices[i]['choice'].get() == 'formula':
                self.pipe.cmd_in(set_str + self.bulk_format[i][0].name
                        + ' = ' + fill_val)
        self.refresh()


    def make_bulk_frame(self, parent, bulk_item, usebmd=False):
        '''
        Creates a frame with all the widgets needed for one parameter in the
        bulk settings window.
        parent: parent widget for this frame
        bulk_item: tao_parameter instance for this frame's parameter
        usebmd: if set true, and option will be given to set to the
                base/model/design/meas/ref value
        Returns a tuple (fill_frame, where_frame, fill_vars)
        '''
        fill_frame = tk.Frame(parent)
        fill_frame.columnconfigure(2, weight=1)
        # Title the column
        tk.Label(fill_frame, text=bulk_item.name).grid(row=0, column=0,
                columnspan=3, sticky='EW')
        # What to fill with
        fill_choice = tk.StringVar()
        fill_choice.set('none')
        fill_vars = {}
        fill_vars['choice'] = fill_choice

        # No change
        none_button = tk.Radiobutton(fill_frame, text="",
                variable=fill_choice, value='none')
        none_button.grid(row=1, column=0, sticky='W')
        tk.Label(fill_frame, text="No Change").grid(
                row=1, column=1, sticky='W')
        # Constant
        const_button = tk.Radiobutton(fill_frame, text="",
                variable=fill_choice, value='const')
        const_button.grid(row=2, column=0, sticky='W')
        tk.Label(fill_frame, text="Constant:").grid(
                row=2, column=1, sticky='W')
        if bulk_item.type == 'REAL':
            const_var = tk.StringVar()
            const_wid = tk.Entry(fill_frame, textvariable=const_var)
            const_wid.grid(row=2, column=2, sticky='EW')
        elif bulk_item.type == 'LOGIC':
            const_var = tk.BooleanVar()
            const_wid = tk.Checkbutton(fill_frame, variable=const_var)
            const_wid.grid(row=2, column=2, sticky='W')
        fill_vars['const'] = const_var
        # Base/Model/Design/Meas/Ref
        if usebmd:
            bmd_var = tk.StringVar()
            bmd_var.set('Base')
            bmd_button = tk.Radiobutton(fill_frame, text="",
                    variable=fill_choice, value='bmd')
            bmd_button.grid(row=3, column=0, sticky='W')
            tk.Label(fill_frame, text="From:").grid(
                    row=3, column=1, sticky='W')
            bmd_wid = tk.OptionMenu(fill_frame, bmd_var, "Base", "Model",
                    "Design", "Measure", "Reference")
            bmd_wid.grid(row=3, column=2, sticky='EW')
            fill_vars['bmd'] = bmd_var
        # Formula
        if bulk_item.type == 'REAL':
            formula_button = tk.Radiobutton(fill_frame, text="",
                    variable=fill_choice, value='formula')
            formula_button.grid(row=4, column=0, sticky='W')
            tk.Label(fill_frame, text="Formula:").grid(
                    row=4, column=1, sticky='W')
            formula_var = tk.StringVar()
            formula_wid = tk.Entry(fill_frame, textvariable=formula_var)
            formula_wid.grid(row=4, column=2)
            fill_vars['formula'] = formula_var

        # Where to fill
        where_frame = tk.Frame(parent)
        fill_where = tk.StringVar()
        fill_where.set('all')
        fill_vars['where'] = fill_where
        fill_range = tk.StringVar()
        # All
        all_button = tk.Radiobutton(where_frame, text="",
                variable=fill_where, value='all')
        all_button.grid(row=0, column=0, sticky='W')
        tk.Label(where_frame, text="All").grid(row=0, column=1, sticky='W')
        # Range
        range_button = tk.Radiobutton(where_frame, text="",
                variable=fill_where, value='range')
        range_button.grid(row=1, column=0, sticky='W')
        tk.Label(where_frame, text="Range:").grid(
                row=1, column=1, sticky='W')
        range_wid = tk.Entry(where_frame, textvariable=fill_range)
        range_wid.grid(row=1, column=2, sticky='EW')
        fill_vars['range'] = fill_range

        return (fill_frame, where_frame, fill_vars)


    def open_detail_window(self, index):
        '''
        Opens up a detail window for the given index.
        '''
        detail_title = self.array_name + '[' + str(index) + ']'
        win = tao_parameter_window(self.root, detail_title, self.param_list, self.pipe, parent=self)

        set_str = self.set_format.format(self.array_name, index)
        b = tk.Button(win.button_frame, text="Apply changes",
                command=lambda : self.detail_set_callback(win.tao_list,set_str))
        b.pack()

    def detail_set_callback(self, tao_list, set_str):
        tao_set(tao_list, set_str, self.pipe)
        self.refresh()

#-----------------------------------------------------
# Branch/element choosing widgets
class tao_branch_widgets:
    '''
    Provides several widgets for slecting universe,
    branch, and elements
    Available widgets:
    self.uni_chooser: OptionMenu to pick the universe index
    self.branch_chooser: OptionMenu to pick the branch (displays name and index)
    self.ele_chooser: ttk Combobox for selecting an element.
            May be specified by name or index
    self.bmd_chooser: OptionMenu for choosing base, model, or design
    self.bme_chooser: OptionMenu for choosing beginning, middle, or end
    In addition, the class provides tk variables for each of these widgets,
            named the same but without _chooser at the end
    '''
    def __init__(self, parent, pipe, default=None):
        '''
        parent: the parent widget where these widgets will be placed
        pipe: tao_interface object
        default: specify the default state of the widgets
        '''
        self.parent = parent
        self.pipe = pipe
        self.default = default

        # Get list of universes, branches, and
        # number of elements in each branch
        n_uni = self.pipe.cmd_in("python super_universe")
        n_uni = n_uni.splitlines()
        n_uni = int(n_uni[0].split(';')[3])
        self.u_list = [] # List of universes
        self.b_list = {} # b_list[i] = branches in universe i
        self.b_name_list = {} # b_name_list[i] = branch names in universe i
        self.e_list = {} #e_list[i][j] = range from 0 to the max element number
        self.e_name_list = {} #e_name_list[i][j] = ele names in branch j of uni i
        self.e_display_list = {} # indexed the same as e_list and e_name_list,
                                                         # but for displaying the number and name
        # There are three separate lists for the elements because
        # users can specify elements by name, by number,
        # or by selecting from a dropdown, which displays
        # the name and number
        for i in range(n_uni):
            self.u_list.append(str(i+1))

        for u in self.u_list:
            self.b_list[u] = self.pipe.cmd_in("python lat_general " + str(u))
            self.b_list[u] = self.b_list[u].splitlines()
            self.b_name_list[u] = []
            self.e_list[u] = {}
            self.e_name_list[u] = {}
            self.e_display_list[u] = {}
            for i in range(len(self.b_list[u])):
                branch_num = self.b_list[u][i].split(';')[0]
                branch_name = self.b_list[u][i].split(';')[1]
                ele_num = int(self.b_list[u][i].split(';')[3])
                self.b_list[u][i] = branch_num
                self.b_name_list[u].append('(' + branch_num + ') ' + branch_name)
                self.e_list[u][branch_num] = range(ele_num+1)
                self.e_display_list[u][branch_num] = []
                ele_names = self.pipe.cmd_in(
                        "python lat_ele_list " + u + '@' + branch_num)
                ele_names = ele_names.splitlines()
                for j in range(len(ele_names)):
                    ele_names[j] = ele_names[j].split(';')[1]
                    display_name = '(' + str(j) + ') ' + ele_names[j]
                    self.e_display_list[u][branch_num].append(display_name)
                self.e_name_list[u][branch_num] = ele_names

        # The variables and widgets
        self.uni = tk.StringVar()
        self.branch = tk.StringVar()
        self.branch_name = tk.StringVar()
        self.ele = tk.StringVar()
        self.ele_label = tk.StringVar() # For showing index range
        self.bmd = tk.StringVar() # Base/Model/Design
        self.bme = tk.StringVar() # Beginning/Middle/End
        if (default != None):
            self.uni.set(str(default[0]))
            self.branch.set(str(default[1]))
            self.ele.set(str(default[2]))
            self.bmd.set(str(default[3]))
        else:
            self.uni.set(str(self.u_list[0]))
            self.branch.set(str(self.b_list[self.uni.get()][0]))
            self.ele.set('0')
            self.bmd.set("Model")
        self.branch_name.set(self.b_name_list[self.uni.get()][0])
        self.branch_name.trace('w', self.update_branch)
        try: # in case self.ele was set by name and not by index
            ele_num = int(self.ele.get())
            self.ele.set(
                    self.e_display_list[self.uni.get()][self.branch.get()][ele_num])
        except ValueError:
            pass
        self.ele_label.set("Element (0 to "
                + str(self.e_list[self.uni.get()][self.branch.get()][-1]) + ")")

        self.uni_chooser = tk.OptionMenu(
                self.parent, self.uni, *self.u_list, command=self.make_branch)
        self.branch_chooser = tk.OptionMenu(self.parent, self.branch_name,
                *self.b_name_list[self.uni.get()], command=self.make_ele_chooser)
        self.ele_chooser = ttk.Combobox(self.parent, textvariable=self.ele,
                values=self.e_display_list[self.uni.get()][self.branch.get()])
        #self.ele_chooser.bind("<<ComboboxSelected>>", self.refresh)
        #self.ele_chooser.bind("<Return>", self.refresh)
        self.bmd_chooser = tk.OptionMenu(
                self.parent, self.bmd, "Base", "Model", "Design")
        self.bme_chooser = tk.OptionMenu(
                self.parent, self.bme, "Beginning", "Middle", "End")
        self.bme.set("End")

    def make_branch(self, event=None):
        '''
        This is necessary because the list of options
        to show in the branch chooser depends on what
        universe you are in
        '''
        # Get the current geometry manager and properties
        manager = self.branch_chooser.winfo_manager()
        if manager == 'pack':
            props = self.branch_chooser.pack_info()
        elif manager == 'grid':
            props = self.branch_chooser.grid_info()
        elif manager == 'place':
            props = self.branch_chooser.place_info()
        else: #No need to remake the branch_chooser
            return

        # Remake the branch_chooser
        self.branch_chooser.destroy()
        self.branch_chooser = tk.OptionMenu(self.parent, self.branch_name,
                *self.b_name_list[self.uni.get()], command=self.make_ele_chooser)
        self.branch.set(self.b_list[self.uni.get()][0])
        self.branch_name.set(self.b_name_list[self.uni.get()][0])
        self.make_ele_chooser()

        # Put the branch_chooser back correctly
        if manager == 'pack':
            self.branch_chooser.pack(**props)
        if manager == 'grid':
            self.branch_chooser.grid(**props)
        if manager == 'place':
            self.branch_chooser.place(**props)

    def make_ele_chooser(self, event=None):
        '''
        This is necessary because different branches
        have different lists of elements to display
        '''
        # Get the current geometry manager and properties
        manager = self.ele_chooser.winfo_manager()
        if manager == 'pack':
            props = self.ele_chooser.pack_info()
        elif manager == 'grid':
            props = self.ele_chooser.grid_info()
        elif manager == 'place':
            props = self.ele_chooser.place_info()
        else: #No need to remake the branch_chooser
            return

        # Update self.branch to match self.branch_name
        ix = self.b_name_list[self.uni.get()].index(self.branch_name.get())
        self.branch.set(self.b_list[self.uni.get()][ix])

        #self.ele_chooser.destroy()
        #self.ele_chooser = ttk.Combobox(self.parent, textvariable=self.ele,
        self.ele_chooser.configure(values=self.e_display_list[self.uni.get()][self.branch.get()])
        #self.ele_chooser.bind("<<ComboboxSelected>>", self.update)
        #self.ele_chooser.bind("<Return>", self.update)
        self.ele_label.set("Element (0 to "
                + str(self.e_list[self.uni.get()][self.branch.get()][-1]) + ")")
        # set self.ele to element 0 in the first branch
        self.ele.set(self.e_display_list[self.uni.get()][self.branch.get()][0])

        # Put the branch_chooser back correctly
        if manager == 'pack':
            self.ele_chooser.pack(**props)
        if manager == 'grid':
            self.ele_chooser.grid(**props)
        if manager == 'place':
            self.ele_chooser.place(**props)

    def update(self, event=None):
        '''
        Sets self.ele to its index value (as a string)
        Returns 0 if setting self.ele failed, 1 if succeeded
        '''
        # Make sure the element field has an actual element in it
        if (self.ele.get() not in
                self.e_name_list[self.uni.get()][self.branch.get()]) \
                & (self.ele.get() not in
                        self.e_display_list[self.uni.get()][self.branch.get()]):
            try:
                if (int(self.ele.get()) not in
                        self.e_list[self.uni.get()][self.branch.get()]):
                    messagebox.showwarning("Error", "Element not found", parent=self.parent)
                    return 0
            except ValueError:
                messagebox.showwarning("Error", "Element not found", parent=self.parent)
                return 0

        # Set self.ele, in case the element
        # was specified by name or display name
        if self.ele.get() in self.e_name_list[self.uni.get()][self.branch.get()]:
            ele_num = self.e_name_list[self.uni.get()][self.branch.get()].index(
                    self.ele.get())
            self.ele.set(str(ele_num))
        elif self.ele.get() in self.e_display_list[self.uni.get()][self.branch.get()]:
            ele_num = self.e_display_list[self.uni.get()][self.branch.get()].index(
                    self.ele.get())
            self.ele.set(str(ele_num))
        return 1

    def update_branch(self, *args):
        '''
        Trace callback for self.branch_name to update self.branch
        '''
        ix = self.b_name_list[self.uni.get()].index(self.branch_name.get())
        self.branch.set(self.b_list[self.uni.get()][ix])

class tao_progress_window(tk.Toplevel):
    '''
    Provides a window for displaying progress bars for long running processes
    Use the grid geometry manager for placing extra widgets in this window
    Row 0 is reserved for external manipulation (e.g. for a label)
    Properties:
    self.label_vars: list of variables used to set labels
    self.ix: counter used for external bookkeeping
    Init args:
    root: the Tk root window
    parent: the parent window for these progress bars
    num_bars: the number of progress bars
    Other init arguments are passed to the tk.Toplevel.__init__ method
    '''
    def __init__(self, root, parent, num_bars, *args, **kwargs):
        tk.Toplevel.__init__(self, parent, class_='Tao', *args, **kwargs)
        self.parent = parent
        self._labels = [] # not intended to be externally manipulated
        self.label_vars = []
        self._bars = []
        self.ix = 0 #used for external bookkeeping
        for i in range(num_bars):
            self.label_vars.append(tk.StringVar())
            self._labels.append(tk.Label(self, textvariable=self.label_vars[-1]))
            self._labels[-1].grid(row=len(self._labels), column=0)
            self._bars.append(ttk.Progressbar(self, mode='determinate'))
            self._bars[-1].grid(row=len(self._bars), column=1)
    def set_max(self, ix, val):
        '''
        Sets the max value for self.bars[ix] to val
        '''
        self._bars[ix].configure(maximum=val)
    def set_val(self, ix, val, update=True):
        '''
        Sets to value of self.bars[ix] to val
        Note that self.set_max should be used first to set the progress bar max
        Use update=False to prevent running self.parent.update_idletaks()
        (usually necessary to make progress bar reflect actual progress)
        '''
        self._bars[ix].configure(value=val)
        if update:
            self.parent.update_idletasks()

class tao_message_box(tk.Toplevel):
    '''
    Custom messageboxes for when tkinter.messagebox does not suffice
    root: the Tk root window
    parent: the parent window for this message box
    tk_var: the tk.StringVar to write the results to
    title: title for this window
    message: the text to be displayed
    choices: list of strings for the user to pick from
    orient: grid the buttons horizontally or vertically (may be 'horizontal' or 'vertical')
    Other arguments are passed to the tk.Toplevel.__init__ method
    '''
    def __init__(self, root, parent, tk_var, title='', message='', choices=[],
            orient='horizontal', *args, **kwargs):
        tk.Toplevel.__init__(self, parent, class_='Tao', *args, **kwargs)
        self.transient(parent)
        if title:
            self.title(title)
        self.tk_var = tk_var
        # grid buttons first to get appropriate window width
        i = 0
        for c in choices:
            if orient == 'horizontal':
                tk.Button(self, text=c, command=self.button_callback(c)).grid(
                        row=1, column=i, sticky='EW', padx=5)
            elif orient == 'vertical':
                tk.Button(self, text=c, command=self.button_callback(c)).grid(
                        row=i+1, column=0, sticky='EW', padx=5)
            i = i+1
        self.update_idletasks()
        tk.Label(self, text=message, wraplength=self.winfo_width()).grid(row=0, column=0, columnspan=i)
        self.protocol("WM_DELETE_WINDOW", self.cancel)
        self.wait_window(self)

    def button_callback(self, choice):
        return lambda: self.button_cmd(choice)

    def button_cmd(self, choice):
        self.tk_var.set(choice)
        self.destroy()

    def cancel(self, *args):
        self.tk_var.set("")


class tabbed_frame(tk.Frame):
    '''
    Provides a frame for showing multiple frames one at a time
    in a tabbed view.
    parent: The parent widget where this frame will be placed
    new_tab_func: returns a new frame to add to a new tab
    self will be passed as the first positional argument and should
    be the only required positional arguments
    new_tab_func should return a tk.Frame with an instance variable
    self.name, a string that should be used for the initial tab name
    new_tab_func will likely be the init method for a tk.Frame subclass
    pipe: the tao_interface in use
    '''
    def __init__(self, parent, new_tab_func):
        tk.Frame.__init__(self, parent)
        self.new_tab_func = new_tab_func
        # Create self.notebook
        self.notebook = ttk.Notebook(self)
        self.notebook.pack(side='top', fill='both', expand=1)

        self.tab_index = 0 # Marks the currently selected tab
        self.tab_list = []

        # New tab button
        self.new_tab_frame = tk.Frame(self.notebook)
        self.notebook.insert('end', self.new_tab_frame)
        self.notebook.tab(len(self.tab_list), text='+')
        self.notebook.bind('<<NotebookTabChanged>>', self._tab_handler)

    def _tab_handler(self, event=None):
        '''
        Handles new tab creation and updates self.tab_index as necessary
        '''
        # Check if the new tab frame has been selected
        if self.notebook.select() == self.new_tab_frame._w:
            # Add new tab at the end
            self.add_tab(len(self.tab_list), self)
            self.notebook.select(len(self.tab_list)-1)
        else:
            # Update self.tab_index
            for i in range(len(self.tab_list)):
                frame = self.tab_list[i]
                if self.notebook.select() == frame._w:
                    self.tab_index = i
                    # Unblock frame's handlers
                    frame.handler_block = False
                    break

    def add_tab(self, ix=None, *args, **kwargs):
        '''
        Adds a new tab at the specified index
        If ix is not specified, the tab is added at the end of the list
        *args and **kwargs are passed to self.new_tab_func after self
        '''
        if ix == None:
            ix = len(self.tab_list)
        if ix < len(self.tab_list):
            self.tab_list = self.tab_list[:ix] + [self.new_tab_func(*args, **kwargs)] + self.tab_list[ix:]
        else:
            ix = len(self.tab_list)
            self.tab_list.append(self.new_tab_func(*args, **kwargs))
        self.notebook.insert(ix, self.tab_list[ix])
        self.notebook.tab(ix, text=self.tab_list[ix].name)

    def remove_tab(self, ix, destroy=False):
        '''
        Removes the tab at the specified position
        Does not destroy the removed frame unless destroy==True
        '''
        # Make sure ix is not too large
        if ix >= len(self.tab_list):
            ix = len(self.tab_list) - 1
        # Prefer the previous index tab
        if ix > 0:
            self.notebook.select(ix-1)
        else:
            self.notebook.select(0)
        # Remove the tab
        frame = self.tab_list.pop(ix)
        self.notebook.forget(ix)
        if destroy:
            frame.destroy()

    def update_name(self, ix=None):
        '''
        Updates the tab title to match the frame name for the
        tab in the specified position
        If ix==None, runs this method for all tabs
        '''
        if ix not in range(len(self.tab_list)):
            for i in range(len(self.tab_list)):
                self.update_name(i)
        else:
            self.notebook.tab(ix, text=self.tab_list[ix].name)

class ele_shape_frame(tk.Frame):
    '''
    Provides a frame for viewing and editing the shapes that are used to
    display lat_layout and floor_plan graphs in tao

    parent: the parent frame/Toplevel where this frame will be placed
    pipe: the tao_interface used to querry/set ele shapes
    which: either "lat_layout" or "floor_plan"
    '''
    def __init__(self, parent, root, pipe, which):
        self.parent = parent
        self.pipe = pipe
        self.root = root
        tk.Frame.__init__(self, parent)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(1, weight=1)
        self.grid_rowconfigure(2, weight=1)
        self.type = which

        # Format string for the shape_set command
        self.shape_set_format = "python shape_set " + self.type + "^^" \
                + "{shape_ix}^^{ele_id}^^{shape}^^{color}^^{size}" \
                + "^^{label}^^{draw}^^{multi}^^{line_width}"

        self.title_text = tk.StringVar()
        self.title = tk.Label(self, textvariable=self.title_text, font=('Sans', 16, 'bold'))
        self.title.grid(row=0, column=0, sticky='EW')

        # Shape table
        self.table_frame = tk.Frame(self)
        self.table_frame.grid(row=1, column=0, sticky='NSEW')
        self.title_list = ["Index", "Ele ID", "Shape", "Color",
                "Size", "Label", "Draw", "Multi", "Line Width"]
        self.keys = ["shape_ix", "ele_id", "shape", "color", "size", "label", "draw", "multi", "line_width"]
        self.shape_table = ttk.Treeview(self.table_frame, columns=self.title_list, show='headings')
        self.shape_list = []
        self.widths = []
        # Column titles
        for title in self.title_list:
            self.shape_table.heading(title, text=title)
            self.shape_table.column(title, stretch=True, anchor='center')
        # Scrollbars
        hbar = ttk.Scrollbar(
                self.table_frame, orient="horizontal", command=self.shape_table.xview)
        vbar = ttk.Scrollbar(
                self.table_frame, orient="vertical", command=self.shape_table.yview)
        self.shape_table.configure(xscrollcommand=hbar.set, yscrollcommand=vbar.set)

        vbar.pack(side="right", fill="y", expand=0)
        hbar.pack(side="bottom", fill='x', expand=0)
        self.shape_table.pack(side="left", fill="both", expand=1)

        # Button Frame
        self.button_frame = tk.Frame(self)
        self.button_frame.grid(row=2, column=0, sticky='NSEW')
        self.move_up_b = tk.Button(self.button_frame, text="Move up",
                command = self.move_up)
        self.move_down_b = tk.Button(self.button_frame, text="Move down",
                command = self.move_down)
        self.dup_b = tk.Button(self.button_frame, text="Duplicate selected row",
                command = self.duplicate_item)
        self.del_b = tk.Button(self.button_frame, text="Delete selected row",
                command = self.delete_item)
        self.new_top_b = tk.Button(self.button_frame, text="New at top",
                command = self.new_at_top)
        self.new_bottom_b = tk.Button(self.button_frame, text="New at bottom",
                command = self.new_at_bottom)
        self.move_up_b.grid(row=0, column=0, sticky='EW')
        self.move_down_b.grid(row=0, column=1, sticky='EW')
        self.dup_b.grid(row=0, column=2, sticky='EW')
        self.del_b.grid(row=1, column=2, sticky='EW')
        self.new_top_b.grid(row=1, column=0, sticky='EW')
        self.new_bottom_b.grid(row=1, column=1, sticky='EW')
        for i in range(3):
            self.button_frame.grid_columnconfigure(i, weight=1)

        self.refresh()

    def refresh(self):
        '''
        Sets the table contents equal to the output of
        python shape_list lat_layout/floor_plan as requested
        '''
        if self.type not in ["lat_layout", "floor_plan"]:
            return
        # Clear existing rows
        for item in self.shape_table.get_children():
            self.shape_table.delete(item)
        self.shape_list = []

        # Fill rows
        ele_shapes = self.pipe.cmd_in('python shape_list ' + self.type)
        ele_shapes = ele_shapes.splitlines()
        for row in ele_shapes:
            row = row.split(';')
            self.shape_list.append(row)
            self.shape_table.insert("", "end", values=row)

        # Set column widths appropriately
        self.width_adjust()

        # Double click to edit shape
        self.shape_table.bind('<Double-Button-1>', self.edit_shape)

        # Set title
        self.title_text.set("Lat Layout Shapes" if self.type=="lat_layout" else "Floor Plan Shapes")

    def width_adjust(self, event=None):
        '''
        Adjusts column widths and total table width to show all contents
        '''
        # Determine column widths
        widths = [0]*len(self.title_list)
        for i in range(len(self.shape_list)):
            for j in range(len(self.shape_list[i])):
                widths[j] = max([widths[j], 15*len(str(self.shape_list[i][j]))])
        # Update column widths and table width
        for j in range(len(widths)):
            widths[j] = max([len(self.title_list[j])*12, widths[j]])
            self.shape_table.column(self.title_list[j], width=widths[j], minwidth=widths[j])
        self.grid_columnconfigure(0, minsize=sum(widths))
        self.widths = widths

    def get_focus_ix(self):
        '''
        Returns the index of the focus row (lowest index = 0, not 1)
        '''
        x = self.shape_table.focus()
        current_row = self.shape_table.item(x)
        current_row = current_row['values']
        if current_row == "":
            return None
        return int(current_row[0]) - 1

    def edit_shape(self, event=None):
        '''
        Opens a window for editing the selected ele_shape
        '''
        # Open new Toplevel window for shape editing
        win = tk.Toplevel(self, class_='Tao')
        win.grid_columnconfigure(0, weight=1)
        widget_frame = tk.Frame(win)
        widget_frame.grid(row=0, column=0, sticky='NSEW')
        button_frame = tk.Frame(win)
        button_frame.grid(row=1, column=0, sticky='NSEW')
        # Get currently selected row
        current_ix = self.get_focus_ix()
        current_row = self.shape_list[current_ix]

        # Create widgets
        title = "ele_shape(" + str(current_ix+1) + ")"
        win.title(title)
        params = []
        def wid_maker(i):
            '''Helper function'''
            return tao_parameter(self.keys[i], types[i-1], "T", current_row[i])
        types = ["STR", "ENUM", "ENUM", "REAL", "ENUM", "LOGIC", "LOGIC", "INT"]
        for i in range(1, len(current_row)):
            params.append(tk_tao_parameter(wid_maker(i), widget_frame, self.pipe,
                    prefix="shape" if types[i-1]=="ENUM" else ""))
            params[i-1].tk_label.grid(row=i, column=0, sticky='W')
            params[i-1].tk_wid.grid(row=i, column=1, sticky='EW')

        # Apply button
        def write_to_table():
            '''
            Writes the contents of this window into the table,
            makes the necessary changes in tao, and closes this window
            '''
            tao_set(params, "set " + self.type + " ele_shape(" + str(current_ix+1) + ")%", self.pipe)
            win.destroy()
            self.refresh()

        b = tk.Button(button_frame, text="Apply", command=write_to_table)
        b.pack()

    def swap_shapes(self, ix1, ix2):
        '''
        Swaps the shapes with indices ix1 and ix2
        (0 based) and refreshes the window
        '''
        # Input validation
        if not (isinstance(ix1, int) and isinstance(ix2, int)):
            return
        if ix1 == ix2:
            return
        if ix1 >= len(self.shape_list) or ix2 >= len(self.shape_list):
            return
        # Swap the shapes
        self.shape_list[ix1][0] = str(ix2+1)
        self.shape_list[ix2][0] = str(ix1+1)
        cmd_str = self.shape_set_format.format(**dict(zip(self.keys, self.shape_list[ix1])))
        self.pipe.cmd_in(cmd_str)
        cmd_str = self.shape_set_format.format(**dict(zip(self.keys, self.shape_list[ix2])))
        self.pipe.cmd_in(cmd_str)
        self.refresh()

    def move_up(self, event=None):
        '''
        Moves the selected item up one position, and shifts
        the others down by one
        '''
        # Get currently selected row
        current_ix = self.get_focus_ix()
        if current_ix == None:
            return
        self.swap_shapes(current_ix, current_ix-1)

    def move_down(self, event=None):
        '''
        Moves the selected item down one position, and shifts
        the others down by one
        '''
        # Get currently selected row
        current_ix = self.get_focus_ix()
        if current_ix == None:
            return
        self.swap_shapes(current_ix, current_ix+1)

    def duplicate_item(self, event=None):
        '''
        Copies the selected shape and adds a duplicate of it to
        the shape table at the next index
        '''
        ix = self.get_focus_ix()
        if ix == None:
            return
        # Add a new row in tao
        self.pipe.cmd_in("python shape_manage " + self.type + " "
                + str(ix+2) + " add")
        # Set the new row to be a copy of the selected row
        self.shape_list[ix][0] = str(ix+2)
        self.pipe.cmd_in(self.shape_set_format.format(**dict(zip(self.keys, self.shape_list[ix]))))
        self.refresh()

    def delete_item(self, event=None):
        '''
        Removes the selected item from the ele_shape table
        and shifts other ele_shapes accordingly
        '''
        ix = self.get_focus_ix()
        if ix==None:
            return
        self.pipe.cmd_in("python shape_manage " + self.type + " " + str(ix+1) + " delete")
        self.refresh()


    def new_at_top(self, event=None):
        '''
        Adds a new empty shape to the top of the ele_list
        and shifts other elements down
        '''
        # Add the shape
        self.pipe.cmd_in("python shape_manage " + self.type + " 1 add")
        # Set the ele_id non-empty to make it show up in the table
        cmd_str = self.shape_set_format.format(shape_ix=1, ele_id="None", shape="box",
                color="black", shape_size=0, label="none", draw="F", multi="F", line_width=0)
        self.pipe.cmd_in(cmd_str)
        # Refresh table
        self.refresh()

    def new_at_bottom(self, event=None):
        '''
        Adds a new empty shape to the bottom of the ele_list
        and shifts other elements down
        '''
        ix = len(self.shape_list)+1
        # Add the shape
        self.pipe.cmd_in("python shape_manage " + self.type
                + " " + str(ix) + " add")
        # Set the ele_id non-empty to make it show up in the table
        cmd_str = self.shape_set_format.format(shape_ix=ix, ele_id="None", shape="box",
                color="black", size=0, label="none", draw="F", multi="F", line_width=0)
        self.pipe.cmd_in(cmd_str)
        # Refresh table
        self.refresh()
